# Copyright 2015 OpenStack Foundation

# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
NSX-V3 Plugin security & Distributed Firewall integration module
"""

from distutils import version

from oslo_log import log
from oslo_utils import excutils

from vmware_nsxlib.v3 import exceptions
from vmware_nsxlib.v3 import nsx_constants as consts
from vmware_nsxlib.v3 import utils


LOG = log.getLogger(__name__)

PORT_SG_SCOPE = 'os-security-group'


class NsxLibNsGroup(utils.NsxLibApiBase):

    def __init__(self, client, nsxlib_config, firewall_section_handler,
                 nsxlib=None):
        self.firewall_section = firewall_section_handler
        super(NsxLibNsGroup, self).__init__(client, nsxlib_config,
                                            nsxlib=nsxlib)

    @property
    def uri_segment(self):
        return 'ns-groups'

    @property
    def resource_type(self):
        return 'NSGroup'

    def update_nsgroup_and_section(self, name, description, logging,
                                   nsgroup_id, section_id,
                                   log_sg_allowed_traffic):
        logging = (log_sg_allowed_traffic or logging)
        rules = self.firewall_section._process_rules_logging_for_update(
            section_id, logging)
        self.update(nsgroup_id, name, description)
        self.firewall_section.update(section_id, name, description,
                                     rules=rules)

    def update_lport_nsgroups(self, lport_id, original_nsgroups,
                              updated_nsgroups):
        """Update the NSgroups that the logical ports belongs to"""
        added = set(updated_nsgroups) - set(original_nsgroups)
        removed = set(original_nsgroups) - set(updated_nsgroups)
        for nsgroup_id in added:
            try:
                self.add_members(
                    nsgroup_id, consts.TARGET_TYPE_LOGICAL_PORT,
                    [lport_id])
            except exceptions.NSGroupIsFull:
                for nsgroup_id in added:
                    # NOTE(roeyc): If the port was not added to the nsgroup
                    # yet, then this request will silently fail.
                    self.remove_member(
                        nsgroup_id, consts.TARGET_TYPE_LOGICAL_PORT,
                        lport_id)
                raise exceptions.SecurityGroupMaximumCapacityReached(
                    sg_id=nsgroup_id)
            except exceptions.ResourceNotFound:
                with excutils.save_and_reraise_exception():
                    LOG.error("NSGroup %s doesn't exists", nsgroup_id)
        for nsgroup_id in removed:
            self.remove_member(
                nsgroup_id, consts.TARGET_TYPE_LOGICAL_PORT, lport_id)

    def get_nsservice(self, resource_type, **properties):
        service = {'resource_type': resource_type}
        service.update(properties)
        return {'service': service}

    def get_nsgroup_complex_expression(self, expressions):
        return {'resource_type': consts.NSGROUP_COMPLEX_EXP,
                'expressions': expressions}

    def get_switch_tag_expression(self, scope, tag):
        return {'resource_type': consts.NSGROUP_TAG_EXP,
                'target_type': consts.TARGET_TYPE_LOGICAL_SWITCH,
                'scope': scope,
                'tag': tag}

    def get_port_tag_expression(self, scope, tag):
        return {'resource_type': consts.NSGROUP_TAG_EXP,
                'target_type': consts.TARGET_TYPE_LOGICAL_PORT,
                'scope': scope,
                'tag': tag}

    def create(self, display_name, description, tags,
               membership_criteria=None, members=None):
        body = {'display_name': display_name,
                'description': description,
                'tags': tags,
                'members': [] if members is None else members}
        if membership_criteria:
            # Allow caller to pass a list of membership criterias.
            # The 'else' block is maintained for backwards compatibility
            # where in a caller might only send a single membership criteria.
            if isinstance(membership_criteria, list):
                body.update({'membership_criteria': membership_criteria})
            else:
                body.update({'membership_criteria': [membership_criteria]})
        return self.client.create(self.get_path(), body)

    def list(self):
        return self.client.list(
            '%s?populate_references=false' % self.get_path()).get(
            'results', [])

    def update(self, nsgroup_id, display_name=None, description=None,
               membership_criteria=None, members=None, tags_update=None):
        nsgroup = {}
        if display_name is not None:
            nsgroup['display_name'] = display_name
        if description is not None:
            nsgroup['description'] = description
        if members is not None:
            nsgroup['members'] = members
        if membership_criteria is not None:
            if isinstance(membership_criteria, list):
                nsgroup['membership_criteria'] = membership_criteria
            else:
                nsgroup['membership_criteria'] = [membership_criteria]
        if tags_update is not None:
            nsgroup['tags_update'] = tags_update
        return self._update_resource(
            self.get_path(nsgroup_id), nsgroup,
            get_params='?populate_references=true',
            retry=True)

    def get_member_expression(self, target_type, target_id):
        return {
            'resource_type': consts.NSGROUP_SIMPLE_EXP,
            'target_property': 'id',
            'target_type': target_type,
            'op': consts.EQUALS,
            'value': target_id}

    def _update_with_members(self, nsgroup_id, members, action):
        members_update = '%s?action=%s' % (self.get_path(nsgroup_id), action)
        return self.client.create(members_update, members)

    def add_members(self, nsgroup_id, target_type, target_ids):
        members = []
        for target_id in target_ids:
            member_expr = self.get_member_expression(
                target_type, target_id)
            members.append(member_expr)
        members = {'members': members}
        try:
            return self._update_with_members(
                nsgroup_id, members, consts.NSGROUP_ADD_MEMBERS)
        except (exceptions.StaleRevision, exceptions.ResourceNotFound):
            raise
        except exceptions.ManagerError:
            # REVISIT(roeyc): A ManagerError might have been raised for a
            # different reason, e.g - NSGroup does not exists.
            LOG.warning("Failed to add %(target_type)s resources "
                        "(%(target_ids)s) to NSGroup %(nsgroup_id)s",
                        {'target_type': target_type,
                         'target_ids': target_ids,
                         'nsgroup_id': nsgroup_id})

            raise exceptions.NSGroupIsFull(nsgroup_id=nsgroup_id)

    def remove_member(self, nsgroup_id, target_type,
                      target_id, verify=False):
        member_expr = self.get_member_expression(
            target_type, target_id)
        members = {'members': [member_expr]}
        try:
            return self._update_with_members(
                nsgroup_id, members, consts.NSGROUP_REMOVE_MEMBERS)
        except exceptions.ManagerError:
            if verify:
                raise exceptions.NSGroupMemberNotFound(member_id=target_id,
                                                       nsgroup_id=nsgroup_id)

    def read(self, nsgroup_id):
        return self.client.get(
            '%s?populate_references=true' % self.get_path(nsgroup_id))

    def delete(self, nsgroup_id):
        resource = '%s?force=true' % nsgroup_id
        return self._delete_with_retry(resource)

    def find_by_display_name(self, display_name):
        found = []
        for resource in self.list():
            if resource['display_name'] == display_name:
                found.append(resource)
        return found


class NsxLibFirewallSection(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'firewall/sections'

    @property
    def resource_type(self):
        return 'FirewallSection'

    def add_member_to_fw_exclude_list(self, target_id, target_type):
        resource = 'firewall/excludelist?action=add_member'
        body = {"target_id": target_id,
                "target_type": target_type}
        self._create_with_retry(resource, body)

    def remove_member_from_fw_exclude_list(self, target_id, target_type):
        resource = ('firewall/excludelist?action=remove_member&object_id=' +
                    target_id)
        self._create_with_retry(resource)

    def get_excludelist(self):
        return self.client.list('firewall/excludelist')

    def _get_direction(self, sg_rule):
        return (
            consts.IN if sg_rule['direction'] == 'ingress'
            else consts.OUT
        )

    def get_nsservice(self, resource_type, **properties):
        service = {'resource_type': resource_type}
        service.update(properties)
        return {'service': service}

    def _decide_service(self, sg_rule):
        l4_protocol = utils.get_l4_protocol_name(sg_rule['protocol'])

        if l4_protocol in [consts.TCP, consts.UDP]:
            # If port_range_min is not specified then we assume all ports are
            # matched, relying on neutron to perform validation.
            if sg_rule['port_range_min'] is None:
                destination_ports = []
            elif sg_rule['port_range_min'] != sg_rule['port_range_max']:
                # NSX API requires a non-empty range (e.g - '22-23')
                destination_ports = ['%(port_range_min)s-%(port_range_max)s'
                                     % sg_rule]
            else:
                destination_ports = ['%(port_range_min)s' % sg_rule]

            return self.get_nsservice(
                consts.L4_PORT_SET_NSSERVICE,
                l4_protocol=l4_protocol,
                source_ports=[],
                destination_ports=destination_ports)
        elif l4_protocol == consts.ICMPV4:
            # Validate the icmp type & code
            icmp_type = sg_rule['port_range_min']
            icmp_code = sg_rule['port_range_max']
            icmp_strict = self.nsxlib.feature_supported(
                consts.FEATURE_ICMP_STRICT)
            utils.validate_icmp_params(icmp_type, icmp_code, icmp_version=4,
                                       strict=icmp_strict)

            return self.get_nsservice(
                consts.ICMP_TYPE_NSSERVICE,
                protocol=l4_protocol,
                icmp_type=icmp_type,
                icmp_code=icmp_code)
        elif l4_protocol is not None:
            return self.get_nsservice(
                consts.IP_PROTOCOL_NSSERVICE,
                protocol_number=l4_protocol)

    def _build(self, display_name, description, applied_tos, tags):
        return {'display_name': display_name,
                'description': description,
                'stateful': True,
                'section_type': consts.FW_SECTION_LAYER3,
                'applied_tos': [self.get_nsgroup_reference(t_id)
                                for t_id in applied_tos],
                'tags': tags}

    def create_empty(self, display_name, description,
                     applied_tos, tags,
                     operation=consts.FW_INSERT_BOTTOM,
                     other_section=None):
        resource = '%s?operation=%s' % (self.uri_segment, operation)
        body = self._build(display_name, description,
                           applied_tos, tags)
        if other_section:
            resource += '&id=%s' % other_section
        return self._create_with_retry(resource, body)

    def create_with_rules(self, display_name, description, applied_tos=None,
                          tags=None, operation=consts.FW_INSERT_BOTTOM,
                          other_section=None, rules=None):
        resource = '%s?operation=%s' % (self.uri_segment, operation)
        body = {
            'display_name': display_name,
            'description': description,
            'stateful': True,
            'section_type': consts.FW_SECTION_LAYER3,
            'applied_tos': applied_tos or [],
            'tags': tags or []
        }
        if rules is not None:
            resource += '&action=create_with_rules'
            body['rules'] = rules
        if other_section:
            resource += '&id=%s' % other_section
        return self._create_with_retry(resource, body)

    def update(self, section_id, display_name=None, description=None,
               applied_tos=None, rules=None, tags_update=None, force=False):
        resource = self.get_path(section_id)
        params = None
        section = {}
        if rules is not None:
            params = '?action=update_with_rules'
            section['rules'] = rules
        if display_name is not None:
            section['display_name'] = display_name
        if description is not None:
            section['description'] = description
        if applied_tos is not None:
            section['applied_tos'] = [self.get_nsgroup_reference(nsg_id)
                                      for nsg_id in applied_tos]
        if tags_update is not None:
            section['tags_update'] = tags_update

        headers = None
        if force:
            # shared sections (like default section) can serve multiple
            # openstack deployments. If some operate under protected
            # identities, force-overwrite is needed.
            # REVISIT(annak): find better solution for shared sections
            headers = {'X-Allow-Overwrite': 'true'}

        if rules is not None:
            return self._update_resource(resource, section,
                                         headers=headers,
                                         create_action=True,
                                         action_params=params,
                                         retry=True)

        elif any(p is not None for p in (display_name, description,
                                         applied_tos, tags_update)):
            return self._update_resource(resource, section,
                                         headers=headers,
                                         action_params=params,
                                         retry=True)

    def list(self):
        return self.client.list(self.get_path()).get('results', [])

    def delete(self, section_id):
        resource = '%s?cascade=true' % section_id
        return self._delete_with_retry(resource)

    def get_nsgroup_reference(self, nsgroup_id):
        return {'target_id': nsgroup_id,
                'target_type': consts.NSGROUP}

    def get_logicalport_reference(self, port_id):
        return {'target_id': port_id,
                'target_type': consts.TARGET_TYPE_LOGICAL_PORT}

    def get_ip_cidr_reference(self, ip_cidr_block, ip_protocol):
        target_type = (consts.TARGET_TYPE_IPV4ADDRESS
                       if ip_protocol == consts.IPV4
                       else consts.TARGET_TYPE_IPV6ADDRESS)
        return {'target_id': ip_cidr_block,
                'target_type': target_type}

    def get_rule_address(self, target_id, display_name=None, is_valid=True,
                         target_type=consts.TARGET_TYPE_IPV4ADDRESS):
        return {'target_display_name': display_name or '',
                'target_id': target_id,
                'is_valid': is_valid,
                'target_type': target_type}

    def get_l4portset_nsservice(self, sources=None, destinations=None,
                                protocol=consts.TCP):
        return {
            'service': {
                'resource_type': 'L4PortSetNSService',
                'source_ports': sources or [],
                'destination_ports': destinations or [],
                'l4_protocol': protocol}
        }

    def get_rule_dict(self, display_name, sources=None, destinations=None,
                      direction=consts.IN_OUT, ip_protocol=consts.IPV4_IPV6,
                      services=None, action=consts.FW_ACTION_ALLOW,
                      logged=False, disabled=False, applied_tos=None,
                      rule_tag=None):
        rule_dict = {'display_name': display_name,
                     'direction': direction,
                     'ip_protocol': ip_protocol,
                     'action': action,
                     'logged': logged,
                     'disabled': disabled,
                     'sources': sources or [],
                     'destinations': destinations or [],
                     'services': services or []}
        if applied_tos is not None:
            rule_dict['applied_tos'] = applied_tos
        if rule_tag is not None:
            rule_dict['rule_tag'] = rule_tag
        return rule_dict

    def add_rule(self, rule, section_id, operation=consts.FW_INSERT_BOTTOM):
        @utils.retry_upon_exception(exceptions.StaleRevision,
                                    max_attempts=self.client.max_attempts)
        def _do_add_rule():
            resource = '%s/rules' % self.get_path(section_id)
            params = '?operation=%s' % operation
            if (version.LooseVersion(self.nsxlib.get_version()) >=
                version.LooseVersion(consts.NSX_VERSION_2_4_0)):
                rule['_revision'] = self.get(section_id)['_revision']
            return self.client.create(resource + params, rule)
        return _do_add_rule()

    def add_rules(self, rules, section_id, operation=consts.FW_INSERT_BOTTOM):
        @utils.retry_upon_exception(exceptions.StaleRevision,
                                    max_attempts=self.client.max_attempts)
        def _do_add_rules():
            resource = '%s/rules' % self.get_path(section_id)
            params = '?action=create_multiple&operation=%s' % operation
            if (version.LooseVersion(self.nsxlib.get_version()) >=
                version.LooseVersion(consts.NSX_VERSION_2_4_0)):
                rev_id = self.get(section_id)['_revision']
                for rule in rules:
                    rule['_revision'] = rev_id
            return self.client.create(resource + params, {'rules': rules})
        return _do_add_rules()

    def delete_rule(self, section_id, rule_id):
        resource = '%s/rules/%s' % (section_id, rule_id)
        return self._delete_with_retry(resource)

    def get_rules(self, section_id):
        resource = '%s/rules' % self.get_path(section_id)
        return self.client.get(resource)

    def get_default_rule(self, section_id):
        rules = self.get_rules(section_id)['results']
        last_rule = rules[-1]
        if last_rule['is_default']:
            return last_rule

    def _get_fw_rule_from_sg_rule(self, sg_rule, nsgroup_id, rmt_nsgroup_id,
                                  logged, action):
        # IPV4 or IPV6
        ip_protocol = sg_rule['ethertype'].upper()
        direction = self._get_direction(sg_rule)

        if sg_rule.get(consts.LOCAL_IP_PREFIX):
            local_ip_prefix = self.get_ip_cidr_reference(
                sg_rule[consts.LOCAL_IP_PREFIX],
                ip_protocol)
        else:
            local_ip_prefix = None

        source = None
        local_group = self.get_nsgroup_reference(nsgroup_id)
        if sg_rule['remote_ip_prefix'] is not None:
            source = self.get_ip_cidr_reference(
                sg_rule['remote_ip_prefix'], ip_protocol)
            destination = local_ip_prefix or local_group
        else:
            if rmt_nsgroup_id:
                source = self.get_nsgroup_reference(rmt_nsgroup_id)
            destination = local_ip_prefix or local_group
        if direction == consts.OUT:
            source, destination = destination, source

        service = self._decide_service(sg_rule)
        name = sg_rule['id']

        return self.get_rule_dict(name, [source] if source else None,
                                  [destination] if destination else None,
                                  direction,
                                  ip_protocol,
                                  [service] if service else None,
                                  action, logged)

    def create_section_rules(self, section_id, nsgroup_id,
                             logging_enabled, action, security_group_rules,
                             ruleid_2_remote_nsgroup_map):
        # 1. translate rules
        # 2. insert in section
        # 3. return the rules
        firewall_rules = []
        for sg_rule in security_group_rules:
            remote_nsgroup_id = ruleid_2_remote_nsgroup_map[sg_rule['id']]
            fw_rule = self._get_fw_rule_from_sg_rule(
                sg_rule, nsgroup_id, remote_nsgroup_id,
                logging_enabled, action)

            firewall_rules.append(fw_rule)
        return self.add_rules(firewall_rules, section_id)

    def set_rule_logging(self, section_id, logging):
        rules = self._process_rules_logging_for_update(
            section_id, logging)
        self.update(section_id, rules=rules)

    def _process_rules_logging_for_update(self, section_id, logging_enabled):
        rules = self.get_rules(section_id).get('results', [])
        update_rules = False
        for rule in rules:
            if rule['logged'] != logging_enabled:
                rule['logged'] = logging_enabled
                update_rules = True
        return rules if update_rules else None

    def init_default(self, name, description, nested_groups,
                     log_sg_blocked_traffic):
        LOG.info("Initializing the default section named %s", name)
        fw_sections = self.list()
        for section in reversed(fw_sections):
            if section['display_name'] == name:
                LOG.info("Found existing default section %s", section['id'])
                break
        else:
            tags = self.build_v3_api_version_tag()
            section = self.create_empty(
                name, description, nested_groups, tags)
            LOG.info("Creating a new default section %s", section['id'])

        block_rule = self.get_rule_dict(
            'Block All', action=consts.FW_ACTION_DROP,
            logged=log_sg_blocked_traffic)
        # TODO(roeyc): Add additional rules to allow IPV6 NDP.
        dhcp_client = self.get_nsservice(
            consts.L4_PORT_SET_NSSERVICE,
            l4_protocol=consts.UDP,
            source_ports=[67],
            destination_ports=[68])
        dhcp_client_rule_in = self.get_rule_dict(
            'DHCP Reply', direction=consts.IN,
            services=[dhcp_client])

        dhcp_server = (
            self.get_nsservice(
                consts.L4_PORT_SET_NSSERVICE,
                l4_protocol=consts.UDP,
                source_ports=[68],
                destination_ports=[67]))
        dhcp_client_rule_out = self.get_rule_dict(
            'DHCP Request', direction=consts.OUT,
            services=[dhcp_server])

        self.update(section['id'],
                    name, section['description'],
                    applied_tos=nested_groups,
                    rules=[dhcp_client_rule_out,
                           dhcp_client_rule_in,
                           block_rule],
                    force=True)
        return section['id']


class NsxLibIPSet(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'ip-sets'

    @property
    def resource_type(self):
        return 'IPSet'

    def create(self, display_name, description=None, ip_addresses=None,
               tags=None):
        body = {
            'display_name': display_name,
            'description': description or '',
            'ip_addresses': ip_addresses or [],
            'tags': tags or []
        }
        return self.client.create(self.get_path(), body)

    def update(self, ip_set_id, display_name=None, description=None,
               ip_addresses=None, tags_update=None, update_payload_cbk=None):
        # The update_payload_cbk function takes two arguments.
        # The first one is the result from the internal GET request.
        # The second one is a dict of user-provided attributes,
        # which can be changed inside the callback function and
        # used as the new payload for the following PUT request.
        # For example, users want to combine the new ip_addresses
        # passed to update() with the original ip_addresses retrieved
        # from the internal GET request instead of overriding the
        # original ip_addresses.
        ip_set = {}
        if tags_update:
            ip_set['tags_update'] = tags_update
        if display_name is not None:
            ip_set['display_name'] = display_name
        if description is not None:
            ip_set['description'] = description
        if ip_addresses is not None:
            ip_set['ip_addresses'] = ip_addresses
        return self._update_resource(self.get_path(ip_set_id),
                                     ip_set, retry=True,
                                     update_payload_cbk=update_payload_cbk)

    def read(self, ip_set_id):
        return self.client.get('ip-sets/%s' % ip_set_id)

    def delete(self, ip_set_id):
        self._delete_with_retry(ip_set_id)

    def get_ipset_reference(self, ip_set_id):
        return {'target_id': ip_set_id,
                'target_type': consts.IP_SET}



# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import copy
from unittest import mock

from oslo_utils import uuidutils

from vmware_nsxlib.tests.unit.v3 import nsxlib_testcase
from vmware_nsxlib.tests.unit.v3 import test_constants
from vmware_nsxlib.v3 import exceptions as nsxlib_exc
from vmware_nsxlib.v3 import nsx_constants
from vmware_nsxlib.v3 import router as nsx_router


class TestRouter(nsxlib_testcase.NsxClientTestCase):

    def test_validate_tier0(self):
        tier0_groups_dict = {}
        tier0_uuid = uuidutils.generate_uuid()
        rtr = {'edge_cluster_id': test_constants.FAKE_EDGE_CLUSTER_ID}
        with mock.patch.object(self.nsxlib.router._router_client, 'get',
                               return_value=rtr),\
            mock.patch.object(
                self.nsxlib.edge_cluster, 'get',
                return_value=test_constants.FAKE_EDGE_CLUSTER):
            self.nsxlib.router.validate_tier0(tier0_groups_dict, tier0_uuid)
            self.assertEqual(
                tier0_groups_dict[tier0_uuid]['edge_cluster_uuid'],
                test_constants.FAKE_EDGE_CLUSTER_ID)
            self.assertEqual(
                tier0_groups_dict[tier0_uuid]['member_index_list'], [0])

    def test_validate_tier0_fail_no_cluster(self):
        tier0_groups_dict = {}
        tier0_uuid = uuidutils.generate_uuid()
        rtr = {'edge_cluster_id': None}
        with mock.patch.object(self.nsxlib.router._router_client, 'get',
                               return_value=rtr):
            self.assertRaises(
                nsxlib_exc.NsxLibInvalidInput,
                self.nsxlib.router.validate_tier0,
                tier0_groups_dict, tier0_uuid)

    def test_validate_tier0_fail_no_members(self):
        tier0_groups_dict = {}
        tier0_uuid = uuidutils.generate_uuid()
        edge_cluster = copy.copy(test_constants.FAKE_EDGE_CLUSTER)
        edge_cluster['members'] = []
        with mock.patch.object(self.nsxlib.router._router_client, 'get'),\
            mock.patch.object(self.nsxlib.edge_cluster, 'get',
                              return_value=edge_cluster):
            self.assertRaises(
                nsxlib_exc.NsxLibInvalidInput,
                self.nsxlib.router.validate_tier0,
                tier0_groups_dict, tier0_uuid)

    def test_add_router_link_port(self):
        tags = [{'scope': 'a', 'tag': 'b'}]
        tier0_uuid = uuidutils.generate_uuid()
        tier1_uuid = uuidutils.generate_uuid()
        with mock.patch.object(self.nsxlib.router._router_port_client,
                               'create') as port_create:
            tier0_link_port = mock.MagicMock()
            tier1_link_port = mock.MagicMock()
            port_create.side_effect = [tier0_link_port, tier1_link_port]
            self.assertEqual(
                (tier0_link_port, tier1_link_port),
                self.nsxlib.router.add_router_link_port(
                    tier1_uuid, tier0_uuid, tags))
            self.assertEqual(port_create.call_count, 2)

    def test_add_router_link_port_fail(self):
        tags = [{'scope': 'a', 'tag': 'b'}]
        tier0_uuid = uuidutils.generate_uuid()
        tier1_uuid = uuidutils.generate_uuid()
        tier0_link_port_id = uuidutils.generate_uuid()
        with mock.patch.object(self.nsxlib.router._router_port_client,
                               'create') as port_create,\
            mock.patch.object(self.nsxlib.router._router_port_client,
                              'delete') as port_delete:
            tier0_link_port = {'id': tier0_link_port_id}
            port_create.side_effect = [tier0_link_port,
                                       nsxlib_exc.ManagerError]
            self.assertRaises(
                nsxlib_exc.ManagerError,
                self.nsxlib.router.add_router_link_port,
                tier1_uuid, tier0_uuid, tags)
            port_delete.assert_called_once_with(tier0_link_port_id)

    def test_remove_router_link_port(self):
        tier1_uuid = uuidutils.generate_uuid()
        with mock.patch.object(
            self.nsxlib.router._router_port_client, 'get_tier1_link_port',
            return_value=test_constants.FAKE_ROUTER_LINKT1_PORT) as port_get,\
            mock.patch.object(self.nsxlib.router._router_port_client,
                              'delete') as port_delete:
            self.nsxlib.router.remove_router_link_port(tier1_uuid)
            self.assertEqual(port_get.call_count, 1)
            self.assertEqual(port_delete.call_count, 2)

    def test_add_centralized_service_port(self):
        logical_router_id = uuidutils.generate_uuid()
        logical_port_id = uuidutils.generate_uuid()
        display_name = mock.Mock()
        tags = mock.Mock()
        address_groups = mock.Mock()
        port = mock.Mock()
        with mock.patch.object(
                self.nsxlib.router._router_port_client, 'create',
                return_value=port) as create_port:
            csp = self.nsxlib.router.add_centralized_service_port(
                logical_router_id, display_name=display_name, tags=tags,
                logical_port_id=logical_port_id, address_groups=address_groups)
            create_port.assert_called_once_with(
                logical_router_id, display_name=display_name, tags=tags,
                logical_port_id=logical_port_id, address_groups=address_groups,
                resource_type=nsx_constants.LROUTERPORT_CENTRALIZED)
            self.assertEqual(csp, port)

    def test_update_advertisement(self):
        router_id = test_constants.FAKE_ROUTER_UUID
        nat = True
        con = False
        static = True
        enabled = True
        vip = False
        snat = True
        data = {'advertise_route_nat': nat,
                'advertise_route_connected': con,
                'advertise_route_static': static,
                'enabled': enabled,
                'advertise_lb_vip': vip,
                'advertise_lb_snat_ip': snat}
        api_client = self.nsxlib.router.nsxlib.logical_router.client
        with mock.patch.object(api_client, 'get', return_value={}),\
            mock.patch.object(api_client, 'update') as client_update,\
            mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
                       return_value=nsxlib_testcase.LATEST_VERSION):
            self.nsxlib.router.update_advertisement(router_id, **data)
            client_update.assert_called_with(
                'logical-routers/%s/routing/advertisement' % router_id,
                {'advertise_nat_routes': nat,
                 'advertise_nsx_connected_routes': con,
                 'advertise_static_routes': static,
                 'enabled': enabled,
                 'advertise_lb_vip': vip,
                 'advertise_lb_snat_ip': snat},
                headers=None)

    def test_update_advertisement_lb_unsupported(self):
        router_id = test_constants.FAKE_ROUTER_UUID
        nat = True
        con = False
        static = True
        enabled = True
        vip = False
        snat = True
        data = {'advertise_route_nat': nat,
                'advertise_route_connected': con,
                'advertise_route_static': static,
                'enabled': enabled,
                'advertise_lb_vip': vip,
                'advertise_lb_snat_ip': snat}
        api_client = self.nsxlib.router.nsxlib.logical_router.client
        with mock.patch.object(api_client, 'get', return_value={}),\
            mock.patch.object(api_client, 'update') as client_update,\
            mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
                       return_value='2.0.0'):
            self.nsxlib.router.update_advertisement(router_id, **data)
            client_update.assert_called_with(
                'logical-routers/%s/routing/advertisement' % router_id,
                {'advertise_nat_routes': nat,
                 'advertise_nsx_connected_routes': con,
                 'advertise_static_routes': static,
                 'enabled': enabled},
                headers=None)

    def test_delete_gw_snat_rule(self):
        logical_router_id = test_constants.FAKE_ROUTER_UUID
        gw_ip = '1.1.1.1'
        with mock.patch.object(self.nsxlib.router.nsxlib.logical_router,
                               'delete_nat_rule_by_values') as del_api:
            self.nsxlib.router.delete_gw_snat_rule(logical_router_id, gw_ip)
            del_api.assert_called_with(
                logical_router_id,
                translated_network=gw_ip)

    def test_delete_gw_snat_rule_by_source(self):
        logical_router_id = test_constants.FAKE_ROUTER_UUID
        gw_ip = '1.1.1.1'
        source_net = '2.0.0.0/24'
        with mock.patch.object(self.nsxlib.router.nsxlib.logical_router,
                               'delete_nat_rule_by_values') as del_api:
            self.nsxlib.router.delete_gw_snat_rule_by_source(
                logical_router_id, gw_ip, source_net)
            del_api.assert_called_with(
                logical_router_id,
                translated_network=gw_ip,
                match_source_network=source_net,
                skip_not_found=False, strict_mode=True)

    def test_delete_gw_snat_rules(self):
        logical_router_id = test_constants.FAKE_ROUTER_UUID
        gw_ip = '1.1.1.1'
        with mock.patch.object(self.nsxlib.router.nsxlib.logical_router,
                               'delete_nat_rule_by_values') as del_api:
            self.nsxlib.router.delete_gw_snat_rules(
                logical_router_id, gw_ip)
            del_api.assert_called_with(
                logical_router_id,
                translated_network=gw_ip,
                skip_not_found=True, strict_mode=False)

    def test_add_gw_snat_rule(self):
        logical_router_id = test_constants.FAKE_ROUTER_UUID
        gw_ip = '1.1.1.1'
        with mock.patch.object(self.nsxlib.router.nsxlib.logical_router,
                               'add_nat_rule') as add_api:
            self.nsxlib.router.add_gw_snat_rule(
                logical_router_id, gw_ip)
            add_api.assert_called_with(
                logical_router_id,
                translated_network=gw_ip,
                action="SNAT",
                bypass_firewall=True,
                source_net=None,
                rule_priority=nsx_router.GW_NAT_PRI,
                tags=None,
                display_name=None)

    def test_update_router_edge_cluster(self):
        logical_router_id = test_constants.FAKE_ROUTER_UUID
        ec_id = test_constants.FAKE_EDGE_CLUSTER
        with mock.patch.object(self.nsxlib.router.nsxlib.logical_router,
                               'update') as update_api:
            self.nsxlib.router.update_router_edge_cluster(
                logical_router_id, ec_id)
            update_api.assert_called_with(
                logical_router_id,
                edge_cluster_id=ec_id)

    def test_update_router_transport_zone(self):
        logical_router_id = test_constants.FAKE_ROUTER_UUID
        tz_id = test_constants.FAKE_TZ_UUID
        with mock.patch.object(self.nsxlib.router.nsxlib.logical_router,
                               'update') as update_api:
            self.nsxlib.router.update_router_transport_zone(
                logical_router_id, tz_id)
            update_api.assert_called_with(
                logical_router_id,
                transport_zone_id=tz_id)

    def test_create_logical_router_intf_port_by_ls_id(self):
        logical_router_id = uuidutils.generate_uuid()
        display_name = 'dummy'
        tags = []
        ls_id = uuidutils.generate_uuid()
        logical_switch_port_id = uuidutils.generate_uuid()
        address_groups = []
        with mock.patch.object(
            self.nsxlib.router._router_port_client,
            "get_by_lswitch_id",
            side_effect=nsxlib_exc.ResourceNotFound()) as get_port,\
            mock.patch.object(self.nsxlib.router._router_port_client,
                              "create") as create_port:
            self.nsxlib.router.create_logical_router_intf_port_by_ls_id(
                logical_router_id,
                display_name,
                tags,
                ls_id,
                logical_switch_port_id,
                address_groups)
            get_port.assert_called_once_with(ls_id)
            create_port.assert_called_once_with(
                logical_router_id, display_name, tags,
                nsx_constants.LROUTERPORT_DOWNLINK,
                logical_switch_port_id, address_groups, urpf_mode=None,
                relay_service_uuid=None)

    def test_add_fip_nat_rules(self):
        with mock.patch.object(self.nsxlib.logical_router,
                               "add_nat_rule") as add_rule:
            self.nsxlib.router.add_fip_nat_rules(
                test_constants.FAKE_ROUTER_UUID,
                '1.1.1.1', '2.2.2.2')
            self.assertEqual(add_rule.call_count, 2)

    def test_delete_fip_nat_rules(self):
        with mock.patch.object(self.nsxlib.logical_router,
                               "delete_nat_rule_by_values") as del_rule:
            self.nsxlib.router.delete_fip_nat_rules(
                test_constants.FAKE_ROUTER_UUID,
                '1.1.1.1', '2.2.2.2')
            self.assertEqual(del_rule.call_count, 2)

    def test_delete_fip_nat_rules_by_int(self):
        with mock.patch.object(self.nsxlib.logical_router,
                               "delete_nat_rule_by_values") as del_rule:
            self.nsxlib.router.delete_fip_nat_rules_by_internal_ip(
                test_constants.FAKE_ROUTER_UUID, '1.1.1.1')
            self.assertEqual(del_rule.call_count, 2)

    def test_add_static_routes(self):
        dest = '1.1.1.0/24'
        nexthop = '2.2.2.2'
        route = {'destination': dest, 'nexthop': nexthop}
        with mock.patch.object(self.nsxlib.logical_router,
                               "add_static_route") as add_route:
            self.nsxlib.router.add_static_routes(
                test_constants.FAKE_ROUTER_UUID, route)
            add_route.assert_called_once_with(
                test_constants.FAKE_ROUTER_UUID,
                dest, nexthop)

    def test_del_static_routes(self):
        dest = '1.1.1.0/24'
        nexthop = '2.2.2.2'
        route = {'destination': dest, 'nexthop': nexthop}
        with mock.patch.object(self.nsxlib.logical_router,
                               "delete_static_route_by_values") as del_route:
            self.nsxlib.router.delete_static_routes(
                test_constants.FAKE_ROUTER_UUID, route)
            del_route.assert_called_once_with(
                test_constants.FAKE_ROUTER_UUID,
                dest_cidr=dest, nexthop=nexthop)

    def test_has_service_router(self):
        logical_router_id = test_constants.FAKE_ROUTER_UUID
        ec_id = test_constants.FAKE_EDGE_CLUSTER
        lr = {'id': logical_router_id,
              'edge_cluster_id': ec_id}
        with mock.patch.object(self.nsxlib.router.nsxlib.logical_router,
                               'get', return_value=lr):
            res = self.nsxlib.router.has_service_router(
                logical_router_id)
            self.assertTrue(res)

    def test_get_tier0_router_tz(self):
        tier0_uuid = uuidutils.generate_uuid()
        self.nsxlib.feature_supported = mock.MagicMock()
        self.nsxlib.feature_supported.return_value = False
        with mock.patch.object(self.nsxlib.router._router_client, 'get',
                               return_value=test_constants.FAKE_TIERO_ROUTER),\
            mock.patch.object(self.nsxlib.edge_cluster, 'get',
                              return_value=test_constants.FAKE_EDGE_CLUSTER),\
            mock.patch.object(self.nsxlib.transport_node, 'get',
                              return_value=test_constants.FAKE_TRANS_NODE):
            tzs = self.nsxlib.router.get_tier0_router_tz(tier0_uuid)
            self.assertEqual(tzs, [test_constants.FAKE_TZ_UUID])

    def test_get_tier0_router_overlay_tz(self):
        tier0_uuid = uuidutils.generate_uuid()
        self.nsxlib.feature_supported = mock.MagicMock()
        self.nsxlib.feature_supported.return_value = False
        with mock.patch.object(self.nsxlib.router._router_client, 'get',
                               return_value=test_constants.FAKE_TIERO_ROUTER),\
            mock.patch.object(self.nsxlib.edge_cluster, 'get',
                              return_value=test_constants.FAKE_EDGE_CLUSTER),\
            mock.patch.object(self.nsxlib.transport_node, 'get',
                              return_value=test_constants.FAKE_TRANS_NODE),\
            mock.patch.object(self.nsxlib.transport_zone, 'get_transport_type',
                              return_value="OVERLAY"):
            tz = self.nsxlib.router.get_tier0_router_overlay_tz(tier0_uuid)
            self.assertEqual(tz, test_constants.FAKE_TZ_UUID)

    def test_get_tier0_router_overlay_tz_via_advanced_config(self):
        tier0_uuid = uuidutils.generate_uuid()
        with mock.patch.object(self.nsxlib.router._router_client, 'get',
                               return_value=test_constants.FAKE_TIERO_ROUTER):
            tz = self.nsxlib.router.get_tier0_router_overlay_tz(tier0_uuid)
            self.assertEqual(tz, test_constants.FAKE_TZ_UUID)

    def test_get_connected_t0_transit_net(self):
        t1_uuid = uuidutils.generate_uuid()
        transit_net = '1.1.1.0'
        link_port = {
            'subnets': [{'ip_addresses': [transit_net],
                         'prefix_length': '31'}]}
        with mock.patch.object(self.nsxlib.router._router_port_client,
                               'get_tier1_link_port',
                               return_value=link_port) as get_port:
            net = self.nsxlib.router.get_connected_t0_transit_net(t1_uuid)
            get_port.assert_called_with(t1_uuid)
            self.assertEqual('%s' % (transit_net), net)



# Copyright (c) 2015 VMware, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from urllib import parse as urlparse

from oslo_serialization import jsonutils
from oslo_utils import uuidutils
import requests

from vmware_nsxlib.v3 import nsx_constants


FAKE_NAME = "fake_name"
DEFAULT_TIER0_ROUTER_UUID = "efad0078-9204-4b46-a2d8-d4dd31ed448f"
NSX_BRIDGE_CLUSTER_NAME = 'default bridge cluster'
FAKE_MANAGER = "fake_manager_ip"


def make_fake_switch(switch_uuid=None, tz_uuid=None, name=FAKE_NAME):
    if not switch_uuid:
        switch_uuid = uuidutils.generate_uuid()
    if not tz_uuid:
        tz_uuid = uuidutils.generate_uuid()

    fake_switch = {
        "id": switch_uuid,
        "display_name": name,
        "resource_type": "LogicalSwitch",
        "address_bindings": [],
        "transport_zone_id": tz_uuid,
        "replication_mode": nsx_constants.MTEP,
        "admin_state": nsx_constants.ADMIN_STATE_UP,
        "vni": 50056,
        "switching_profile_ids": [
            {
                "value": "64814784-7896-3901-9741-badeff705639",
                "key": "IpDiscoverySwitchingProfile"
            },
            {
                "value": "fad98876-d7ff-11e4-b9d6-1681e6b88ec1",
                "key": "SpoofGuardSwitchingProfile"
            },
            {
                "value": "93b4b7e8-f116-415d-a50c-3364611b5d09",
                "key": "PortMirroringSwitchingProfile"
            },
            {
                "value": "fbc4fb17-83d9-4b53-a286-ccdf04301888",
                "key": "SwitchSecuritySwitchingProfile"
            },
            {
                "value": "f313290b-eba8-4262-bd93-fab5026e9495",
                "key": "QosSwitchingProfile"
            }
        ],
    }
    return fake_switch


def make_fake_dhcp_profile():
    return {"id": uuidutils.generate_uuid(),
            "edge_cluster_id": uuidutils.generate_uuid(),
            "edge_cluster_member_indexes": [0, 1]}


def make_fake_metadata_proxy():
    return {"id": uuidutils.generate_uuid(),
            "metadata_server_url": "http://1.2.3.4",
            "secret": "my secret",
            "edge_cluster_id": uuidutils.generate_uuid(),
            "edge_cluster_member_indexes": [0, 1]}


class MockRequestsResponse(object):
    def __init__(self, status_code, content=None):
        self.status_code = status_code
        self.content = content

    def json(self):
        return jsonutils.loads(self.content)

    def __bool__(self):
        """Returns True if :attr:`status_code` is less than 400.

        This attribute checks if the status code of the response is between
        400 and 600 to see if there was a client error or a server error. If
        the status code, is between 200 and 400, this will return True. This
        is **not** a check to see if the response code is ``200 OK``.
        """
        return self.status_code < 400


class MockRequestSessionApi(object):

    def __init__(self):
        self._store = {}

    def _format_uri(self, uri):
        uri = urlparse.urlparse(uri).path
        while uri.endswith('/'):
            uri = uri[:-1]
        while uri.startswith('/'):
            uri = uri[1:]
        if not self._is_uuid_uri(uri):
            uri = "%s/" % uri
        return uri

    def _is_uuid_uri(self, uri):
        return uuidutils.is_uuid_like(
            urlparse.urlparse(uri).path.split('/')[-1])

    def _query(self, search_key, copy=True):
        items = []
        for uri, obj in self._store.items():
            if uri.startswith(search_key):
                items.append(obj.copy() if copy else obj)
        return items

    def _build_response(self, url, content=None,
                        status=requests.codes.ok, **kwargs):
        if isinstance(content, list):
            content = {
                'result_count': len(content),
                'results': content
            }

        if (content is not None and kwargs.get('headers', {}).get(
                'Content-Type') == 'application/json'):
            content = jsonutils.dumps(content)

        return MockRequestsResponse(status, content=content)

    def _get_content(self, **kwargs):
        content = kwargs.get('data', None)
        if content and kwargs.get('headers', {}).get(
                'Content-Type') == 'application/json':
            content = jsonutils.loads(content)
        return content

    def get(self, url, **kwargs):
        url = self._format_uri(url)

        if self._is_uuid_uri(url):
            item = self._store.get(url)
            code = requests.codes.ok if item else requests.codes.not_found
            return self._build_response(
                url, content=item, status=code, **kwargs)

        return self._build_response(
            url, content=self._query(url), status=requests.codes.ok, **kwargs)

    def _create(self, url, content, **kwargs):
        resource_id = content.get('id')
        if resource_id and self._store.get("%s%s" % (url, resource_id)):
            return self._build_response(
                url, content=None, status=requests.codes.bad, **kwargs)

        resource_id = resource_id or uuidutils.generate_uuid()
        content['id'] = resource_id

        self._store["%s%s" % (url, resource_id)] = content.copy()
        return content

    def post(self, url, **kwargs):
        parsed_url = urlparse.urlparse(url)
        url = self._format_uri(url)

        if self._is_uuid_uri(url):
            if self._store.get(url) is None:
                return self._build_response(
                    url, content=None, status=requests.codes.bad, **kwargs)

        body = self._get_content(**kwargs)
        if body is None:
            return self._build_response(
                url, content=None, status=requests.codes.bad, **kwargs)

        response_content = None

        url_queries = urlparse.parse_qs(parsed_url.query)
        if 'create_multiple' in url_queries.get('action', []):
            response_content = {}
            for resource_name, resource_body in body.items():
                for new_resource in resource_body:
                    created_resource = self._create(
                        url, new_resource, **kwargs)
                    if response_content.get(resource_name, None) is None:
                        response_content[resource_name] = []
                    response_content[resource_name].append(created_resource)
        else:
            response_content = self._create(url, body, **kwargs)

        if isinstance(response_content, MockRequestsResponse):
            return response_content

        return self._build_response(
            url, content=response_content,
            status=requests.codes.created, **kwargs)

    def put(self, url, **kwargs):
        url = self._format_uri(url)

        item = {}
        if self._is_uuid_uri(url):
            item = self._store.get(url, None)
            if item is None:
                return self._build_response(
                    url, content=None,
                    status=requests.codes.not_found, **kwargs)

        body = self._get_content(**kwargs)
        if body is None:
            return self._build_response(
                url, content=None, status=requests.codes.bad, **kwargs)

        item.update(body)
        self._store[url] = item
        return self._build_response(
            url, content=item, status=requests.codes.ok, **kwargs)

    def delete(self, url, **kwargs):
        url = self._format_uri(url)

        if not self._store.get(url):
            return self._build_response(
                url, content=None, status=requests.codes.not_found, **kwargs)

        del self._store[url]
        return self._build_response(
            url, content=None, status=requests.codes.ok, **kwargs)



# Copyright 2015 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import netaddr

from oslo_log import log
from oslo_log import versionutils
import requests

from vmware_nsxlib._i18n import _
from vmware_nsxlib.v3 import core_resources
from vmware_nsxlib.v3 import exceptions
from vmware_nsxlib.v3 import nsx_constants
from vmware_nsxlib.v3 import utils

LOG = log.getLogger(__name__)

# TODO(asarfaty): keeping this for backwards compatibility.
# core_resources.SwitchingProfileTypeId and
# core_resources.PacketAddressClassifier should be used.
# This code will be removed in the future.
SwitchingProfileTypeId = core_resources.SwitchingProfileTypeId
PacketAddressClassifier = core_resources.PacketAddressClassifier


class SwitchingProfileTypes(core_resources.SwitchingProfileTypes):
    # TODO(asarfaty): keeping this for backwards compatibility.
    # This code will be removed in the future.
    def __init__(self):
        versionutils.report_deprecated_feature(
            LOG,
            'resources.SwitchingProfileTypes is deprecated. '
            'Please use core_resources.SwitchingProfileTypes instead.')


class WhiteListAddressTypes(core_resources.WhiteListAddressTypes):
    # TODO(asarfaty): keeping this for backwards compatibility.
    # This code will be removed in the future.
    def __init__(self):
        versionutils.report_deprecated_feature(
            LOG,
            'resources.WhiteListAddressTypes is deprecated. '
            'Please use core_resources.WhiteListAddressTypes instead.')


class SwitchingProfile(core_resources.NsxLibSwitchingProfile):
    # TODO(asarfaty): keeping this for backwards compatibility.
    # This code will be removed in the future.
    def __init__(self, rest_client, *args, **kwargs):
        versionutils.report_deprecated_feature(
            LOG,
            'resources.SwitchingProfile is deprecated. '
            'Please use core_resources.NsxLibSwitchingProfile instead.')
        super(SwitchingProfile, self).__init__(rest_client)


class LogicalPort(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'logical-ports'

    @property
    def resource_type(self):
        return 'LogicalPort'

    def _build_body_attrs(
            self, display_name=None,
            admin_state=True, tags=None,
            address_bindings=None,
            switch_profile_ids=None,
            attachment=None,
            description=None,
            extra_configs=None):
        tags = tags or []
        switch_profile_ids = switch_profile_ids or []
        body = {}
        if tags:
            body['tags'] = tags
        if display_name is not None:
            body['display_name'] = display_name

        if admin_state is not None:
            if admin_state:
                body['admin_state'] = nsx_constants.ADMIN_STATE_UP
            else:
                body['admin_state'] = nsx_constants.ADMIN_STATE_DOWN

        if address_bindings:
            bindings = []
            for binding in address_bindings:
                address_classifier = {
                    'ip_address': binding.ip_address,
                    'mac_address': binding.mac_address
                }
                if binding.vlan is not None:
                    address_classifier['vlan'] = int(binding.vlan)
                bindings.append(address_classifier)
            body['address_bindings'] = bindings
        elif address_bindings is not None:
            body['address_bindings'] = []

        if switch_profile_ids:
            profiles = []
            for profile in switch_profile_ids:
                profiles.append({
                    'value': profile.profile_id,
                    'key': profile.profile_type
                })
            body['switching_profile_ids'] = profiles

        # Note that attachment could be None, meaning reset it.
        if attachment is not False:
            body['attachment'] = attachment

        if description is not None:
            body['description'] = description

        if extra_configs:
            body['extra_configs'] = extra_configs

        return body

    def _prepare_attachment(self, attachment_type, vif_uuid,
                            allocate_addresses, vif_type,
                            parent_vif_id, traffic_tag, app_id, tn_uuid):
        if attachment_type and vif_uuid:
            attachment = {'attachment_type': attachment_type,
                          'id': vif_uuid}
            if vif_type:
                context = {'resource_type': nsx_constants.VIF_RESOURCE_TYPE,
                           'allocate_addresses': allocate_addresses,
                           'vif_type': vif_type}
                if parent_vif_id:
                    context['parent_vif_id'] = parent_vif_id
                    context['traffic_tag'] = traffic_tag
                    context['app_id'] = app_id
                elif tn_uuid:
                    context['transport_node_uuid'] = tn_uuid
                    context['app_id'] = app_id
                attachment['context'] = context
            return attachment
        elif attachment_type is None and vif_uuid is None:
            return None   # reset attachment
        else:
            return False  # no attachment change

    def create(self, lswitch_id, vif_uuid, tags=None,
               attachment_type=nsx_constants.ATTACHMENT_VIF,
               admin_state=True, name=None, address_bindings=None,
               parent_vif_id=None, traffic_tag=None,
               switch_profile_ids=None, vif_type=None, app_id=None,
               allocate_addresses=nsx_constants.ALLOCATE_ADDRESS_NONE,
               description=None, tn_uuid=None,
               extra_configs=None):
        tags = tags or []
        body = {'logical_switch_id': lswitch_id}
        # NOTE(arosen): If parent_vif_id is specified we need to use
        # CIF attachment type.
        attachment = self._prepare_attachment(attachment_type, vif_uuid,
                                              allocate_addresses, vif_type,
                                              parent_vif_id, traffic_tag,
                                              app_id, tn_uuid)
        body.update(self._build_body_attrs(
            display_name=name,
            admin_state=admin_state, tags=tags,
            address_bindings=address_bindings,
            switch_profile_ids=switch_profile_ids,
            attachment=attachment,
            description=description,
            extra_configs=extra_configs))
        return self.client.create(self.get_path(), body=body)

    def delete(self, lport_id):
        self._delete_with_retry('%s?detach=true' % lport_id)

    def update(self, lport_id, vif_uuid,
               name=None, admin_state=None,
               address_bindings=None, switch_profile_ids=None,
               tags_update=None, tags=None,
               attachment_type=nsx_constants.ATTACHMENT_VIF,
               parent_vif_id=None, traffic_tag=None,
               vif_type=None, app_id=None,
               allocate_addresses=nsx_constants.ALLOCATE_ADDRESS_NONE,
               description=None, tn_uuid=None,
               extra_configs=None, force=False):
        # Do not allow tags & tags_update at the same call
        if tags_update and tags:
            raise exceptions.ManagerError(
                details=_("Can't support updating logical port %s both with "
                          "tags and tags_update attributes") % lport_id)

        attachment = self._prepare_attachment(attachment_type, vif_uuid,
                                              allocate_addresses, vif_type,
                                              parent_vif_id, traffic_tag,
                                              app_id, tn_uuid)
        lport = {}
        if tags_update is not None:
            lport['tags_update'] = tags_update
        lport.update(self._build_body_attrs(
            display_name=name,
            admin_state=admin_state,
            address_bindings=address_bindings,
            switch_profile_ids=switch_profile_ids,
            attachment=attachment,
            description=description,
            extra_configs=extra_configs,
            tags=tags))

        headers = None
        if force:
            headers = {'X-Allow-Overwrite': 'true'}

        return self._update_resource(
            self.get_path(lport_id), lport, headers=headers, retry=True)

    def get_by_attachment(self, attachment_type, attachment_id):
        """Return all logical port matching the attachment type and Id"""
        url_suffix = ('?attachment_type=%s&attachment_id=%s' %
                      (attachment_type, attachment_id))
        return self.client.get(self.get_path(url_suffix))

    def get_by_logical_switch(self, logical_switch_id):
        """Return all logical port of a logical switch"""
        url_suffix = '?logical_switch_id=%s' % logical_switch_id
        return self.client.get(self.get_path(url_suffix))


class LogicalRouter(core_resources.NsxLibLogicalRouter):
    # TODO(asarfaty): keeping this for backwards compatibility.
    # This code will be removed in the future.
    def __init__(self, rest_client, *args, **kwargs):
        versionutils.report_deprecated_feature(
            LOG,
            'resources.LogicalRouter is deprecated. '
            'Please use core_resources.NsxLibLogicalRouter instead.')
        super(LogicalRouter, self).__init__(rest_client)


class LogicalRouterPort(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'logical-router-ports'

    @staticmethod
    def _get_relay_binding(relay_service_uuid):
        return {'service_id': {'target_type': 'LogicalService',
                               'target_id': relay_service_uuid}}

    def create(self, logical_router_id,
               display_name,
               tags,
               resource_type,
               logical_port_id,
               address_groups,
               edge_cluster_member_index=None,
               urpf_mode=None,
               relay_service_uuid=None):
        body = {'display_name': display_name,
                'resource_type': resource_type,
                'logical_router_id': logical_router_id,
                'tags': tags or []}
        if address_groups:
            body['subnets'] = address_groups
        if resource_type in [nsx_constants.LROUTERPORT_UPLINK,
                             nsx_constants.LROUTERPORT_DOWNLINK,
                             nsx_constants.LROUTERPORT_CENTRALIZED]:
            body['linked_logical_switch_port_id'] = {
                'target_id': logical_port_id}
        elif resource_type == nsx_constants.LROUTERPORT_LINKONTIER1:
            body['linked_logical_router_port_id'] = {
                'target_id': logical_port_id}
        elif logical_port_id:
            body['linked_logical_router_port_id'] = logical_port_id
        if edge_cluster_member_index:
            body['edge_cluster_member_index'] = edge_cluster_member_index
        if urpf_mode:
            body['urpf_mode'] = urpf_mode
        if relay_service_uuid:
            if (self.nsxlib and
                self.nsxlib.feature_supported(
                    nsx_constants.FEATURE_DHCP_RELAY)):
                body['service_bindings'] = [self._get_relay_binding(
                    relay_service_uuid)]
            else:
                LOG.error("Ignoring relay_service_uuid for router %s port: "
                          "This feature is not supported.", logical_router_id)

        return self.client.create(self.get_path(), body=body)

    def update(self, logical_port_id, **kwargs):
        logical_router_port = {}
        # special treatment for updating/removing the relay service
        if 'relay_service_uuid' in kwargs:
            if kwargs['relay_service_uuid']:
                if (self.nsxlib and
                    self.nsxlib.feature_supported(
                        nsx_constants.FEATURE_DHCP_RELAY)):
                    logical_router_port['service_bindings'] = [
                        self._get_relay_binding(
                            kwargs['relay_service_uuid'])]
                else:
                    LOG.error("Ignoring relay_service_uuid for router "
                              "port %s: This feature is not supported.",
                              logical_port_id)
            else:
                # delete the current one
                if 'service_bindings' in logical_router_port:
                    logical_router_port['service_bindings'] = []
            del kwargs['relay_service_uuid']

        for k in kwargs:
            logical_router_port[k] = kwargs[k]

        return self._update_resource(
            self.get_path(logical_port_id), logical_router_port, retry=True)

    def delete(self, logical_port_id):
        self._delete_with_retry(logical_port_id)

    def get_by_lswitch_id(self, logical_switch_id):
        resource = '?logical_switch_id=%s' % logical_switch_id
        router_ports = self.client.url_get(self.get_path(resource))
        result_count = int(router_ports.get('result_count', "0"))
        if result_count >= 2:
            raise exceptions.ManagerError(
                details=_("Can't support more than one logical router ports "
                          "on same logical switch %s ") % logical_switch_id)
        elif result_count == 1:
            return router_ports['results'][0]
        else:
            err_msg = (_("Logical router link port not found on logical "
                         "switch %s") % logical_switch_id)
            raise exceptions.ResourceNotFound(
                manager=self.client.nsx_api_managers,
                operation=err_msg)

    def update_by_lswitch_id(self, logical_router_id, ls_id, **payload):
        port = self.get_by_lswitch_id(ls_id)
        return self.update(port['id'], **payload)

    def delete_by_lswitch_id(self, ls_id):
        port = self.get_by_lswitch_id(ls_id)
        self.delete(port['id'])

    def get_by_router_id(self, logical_router_id):
        resource = '?logical_router_id=%s' % logical_router_id
        logical_router_ports = self.client.url_get(self.get_path(resource))
        return logical_router_ports['results']

    def get_tier1_link_port(self, logical_router_id):
        logical_router_ports = self.get_by_router_id(logical_router_id)
        for port in logical_router_ports:
            if port['resource_type'] == nsx_constants.LROUTERPORT_LINKONTIER1:
                return port
        raise exceptions.ResourceNotFound(
            manager=self.client.nsx_api_managers,
            operation="get router link port")

    def get_tier0_uplink_ports(self, logical_router_id):
        logical_router_ports = self.get_by_router_id(logical_router_id)
        ports = []
        for port in logical_router_ports:
            if port['resource_type'] == nsx_constants.LROUTERPORT_UPLINK:
                ports.append(port)
        return ports

    def get_tier0_uplink_port(self, logical_router_id):
        ports = self.get_tier0_uplink_ports(logical_router_id)
        if ports:
            return ports[0]

    def get_tier0_uplink_subnets(self, logical_router_id):
        port = self.get_tier0_uplink_port(logical_router_id)
        if port:
            return port.get('subnets', [])
        return []

    def get_tier0_uplink_cidrs(self, logical_router_id):
        # return a list of tier0 uplink ip/prefix addresses
        subnets = self.get_tier0_uplink_subnets(logical_router_id)
        cidrs = []
        for subnet in subnets:
            for ip_address in subnet.get('ip_addresses'):
                cidrs.append('%s/%s' % (ip_address,
                                        subnet.get('prefix_length')))
        return cidrs

    def get_tier0_uplink_ips(self, logical_router_id):
        # return a list of tier0 uplink ip addresses
        subnets = self.get_tier0_uplink_subnets(logical_router_id)
        ips = []
        for subnet in subnets:
            for ip_address in subnet.get('ip_addresses'):
                ips.append(ip_address)
        return ips


class MetaDataProxy(core_resources.NsxLibMetadataProxy):
    # TODO(asarfaty): keeping this for backwards compatibility.
    # This code will be removed in the future.
    def __init__(self, rest_client, *args, **kwargs):
        versionutils.report_deprecated_feature(
            LOG,
            'resources.MetaDataProxy is deprecated. '
            'Please use core_resources.NsxLibMetadataProxy instead.')
        super(MetaDataProxy, self).__init__(rest_client)


class DhcpProfile(core_resources.NsxLibDhcpProfile):
    # TODO(asarfaty): keeping this for backwards compatibility.
    # This code will be removed in the future.
    def __init__(self, rest_client, *args, **kwargs):
        versionutils.report_deprecated_feature(
            LOG,
            'resources.DhcpProfile is deprecated. '
            'Please use core_resources.NsxLibDhcpProfile instead.')
        super(DhcpProfile, self).__init__(rest_client)


class LogicalDhcpServer(utils.NsxLibApiBase):

    def get_dhcp_opt_code(self, name):
        return utils.get_dhcp_opt_code(name)

    @property
    def uri_segment(self):
        return 'dhcp/servers'

    @property
    def resource_type(self):
        return 'LogicalDhcpServer'

    def _construct_server(self, body, dhcp_profile_id=None, server_ip=None,
                          name=None, dns_nameservers=None, domain_name=None,
                          gateway_ip=False, options=None, tags=None):
        if name:
            body['display_name'] = name
        if dhcp_profile_id:
            body['dhcp_profile_id'] = dhcp_profile_id
        if server_ip:
            body['ipv4_dhcp_server']['dhcp_server_ip'] = server_ip
        if dns_nameservers is not None:
            # Note that [] is valid for dns_nameservers, means deleting it.
            body['ipv4_dhcp_server']['dns_nameservers'] = dns_nameservers
        if domain_name:
            body['ipv4_dhcp_server']['domain_name'] = domain_name
        if gateway_ip is not False:
            # Note that None is valid for gateway_ip, means deleting it.
            body['ipv4_dhcp_server']['gateway_ip'] = gateway_ip
        if options:
            body['ipv4_dhcp_server']['options'] = options
        if tags:
            body['tags'] = tags

    def create(self, dhcp_profile_id, server_ip, name=None,
               dns_nameservers=None, domain_name=None, gateway_ip=False,
               options=None, tags=None):
        body = {'ipv4_dhcp_server': {}}
        self._construct_server(body, dhcp_profile_id, server_ip, name,
                               dns_nameservers, domain_name, gateway_ip,
                               options, tags)
        return self.client.create(self.get_path(), body=body)

    def update(self, uuid, dhcp_profile_id=None, server_ip=None, name=None,
               dns_nameservers=None, domain_name=None, gateway_ip=False,
               options=None, tags=None):
        body = {'ipv4_dhcp_server': {}}
        self._construct_server(body, dhcp_profile_id, server_ip, name,
                               dns_nameservers, domain_name, gateway_ip,
                               options, tags)
        return self._update_with_retry(uuid, body)

    def create_binding(self, server_uuid, mac, ip, hostname=None,
                       lease_time=None, options=None, gateway_ip=False):
        body = {'mac_address': mac, 'ip_address': ip}
        if hostname:
            body['host_name'] = hostname
        if lease_time:
            body['lease_time'] = lease_time
        if options:
            body['options'] = options
        if gateway_ip is not False:
            # Note that None is valid for gateway_ip, means deleting it.
            body['gateway_ip'] = gateway_ip
        url = "%s/static-bindings" % server_uuid
        return self.client.url_post(self.get_path(url), body)

    def get_binding(self, server_uuid, binding_uuid):
        url = "%s/static-bindings/%s" % (server_uuid, binding_uuid)
        return self.get(url)

    def update_binding(self, server_uuid, binding_uuid, **kwargs):
        body = {}
        body.update(kwargs)
        url = "%s/static-bindings/%s" % (server_uuid, binding_uuid)
        self._update_resource(self.get_path(url), body, retry=True)

    def delete_binding(self, server_uuid, binding_uuid):
        url = "%s/static-bindings/%s" % (server_uuid, binding_uuid)
        return self.delete(url)


class IpPool(utils.NsxLibApiBase):
    @property
    def uri_segment(self):
        return 'pools/ip-pools'

    @property
    def resource_type(self):
        return 'IpPool'

    def _generate_ranges(self, cidr, gateway_ip):
        """Create list of ranges from the given cidr.

        Ignore the gateway_ip, if defined
        """
        ip_set = netaddr.IPSet(netaddr.IPNetwork(cidr))
        if gateway_ip:
            ip_set.remove(gateway_ip)
        return [{"start": str(r[0]),
                 "end": str(r[-1])} for r in ip_set.iter_ipranges()]

    def create(self, cidr, allocation_ranges=None, display_name=None,
               description=None, gateway_ip=None, dns_nameservers=None,
               tags=None):
        """Create an IpPool.

        Arguments:
        cidr: (required)
        allocation_ranges: (optional) a list of dictionaries, each with
           'start' and 'end' keys, and IP values.
            If None: the cidr will be used to create the ranges,
            excluding the gateway.
        display_name: (optional)
        description: (optional)
        gateway_ip: (optional)
        dns_nameservers: (optional) list of addresses
        """
        if not cidr:
            raise exceptions.InvalidInput(operation="IP Pool create",
                                          arg_name="cidr", arg_val=cidr)
        if not allocation_ranges:
            # generate ranges from (cidr - gateway)
            allocation_ranges = self._generate_ranges(cidr, gateway_ip)

        subnet = {"allocation_ranges": allocation_ranges,
                  "cidr": cidr}
        if gateway_ip:
            subnet["gateway_ip"] = gateway_ip
        if dns_nameservers:
            subnet["dns_nameservers"] = dns_nameservers

        body = {"subnets": [subnet]}
        if description:
            body["description"] = description
        if display_name:
            body["display_name"] = display_name
        if tags:
            body['tags'] = tags

        return self.client.create(self.get_path(), body=body)

    def delete(self, pool_id, force=False):
        url = pool_id
        if force:
            url += '?force=%s' % force
        return self._delete_by_path_with_retry(self.get_path(url))

    def _update_param_in_pool(self, args_dict, key, pool_data):
        # update the arg only if it exists in the args dictionary
        if key in args_dict:
            if args_dict[key]:
                pool_data[key] = args_dict[key]
            else:
                # remove the current value
                del pool_data[key]

    def update(self, pool_id, **kwargs):
        """Update the given attributes in the current pool configuration."""
        # Get the current pool, and remove irrelevant fields
        pool = self.get(pool_id)
        for key in ["resource_type", "_create_time", "_create_user"
                    "_last_modified_user", "_last_modified_time"]:
            pool.pop(key, None)

        # update only the attributes in kwargs
        self._update_param_in_pool(kwargs, 'display_name', pool)
        self._update_param_in_pool(kwargs, 'description', pool)
        self._update_param_in_pool(kwargs, 'tags', pool)
        self._update_param_in_pool(kwargs, 'gateway_ip',
                                   pool["subnets"][0])
        self._update_param_in_pool(kwargs, 'dns_nameservers',
                                   pool["subnets"][0])
        self._update_param_in_pool(kwargs, 'allocation_ranges',
                                   pool["subnets"][0])
        self._update_param_in_pool(kwargs, 'cidr',
                                   pool["subnets"][0])
        return self.client.update(self.get_path(pool_id), pool)

    def allocate(self, pool_id, ip_addr=None, display_name=None, tags=None):
        """Allocate an IP from a pool."""
        # Note: Currently the backend does not support allocation of a
        # specific IP, so an exception will be raised by the backend.
        # Depending on the backend version, this may be allowed in the future
        url = "%s?action=ALLOCATE" % pool_id
        body = {"allocation_id": ip_addr}
        if tags is not None:
            body['tags'] = tags
        if display_name is not None:
            body['display_name'] = display_name
        return self.client.url_post(self.get_path(url), body=body)

    def release(self, pool_id, ip_addr):
        """Release an IP back to a pool."""
        url = "%s?action=RELEASE" % pool_id
        body = {"allocation_id": ip_addr}
        return self.client.url_post(self.get_path(url), body=body)

    def get_allocations(self, pool_id):
        """Return information about the allocated IPs in the pool."""
        url = "%s/allocations" % pool_id
        return self.client.url_get(self.get_path(url))


class NodeHttpServiceProperties(utils.NsxLibApiBase):
    @property
    def uri_segment(self):
        return 'node/services/http'

    @property
    def resource_type(self):
        return 'NodeHttpServiceProperties'

    def get_properties(self):
        return self.client.get(self.get_path())

    def get_rate_limit(self):
        if (self.nsxlib and
            not self.nsxlib.feature_supported(
                nsx_constants.FEATURE_RATE_LIMIT)):
            msg = (_("Rate limit is not supported by NSX version %s") %
                   self.nsxlib.get_version())
            raise exceptions.ManagerError(details=msg)

        properties = self.get_properties()
        return properties.get('service_properties', {}).get(
            'client_api_rate_limit')

    def update_rate_limit(self, value):
        """update the NSX rate limit. default value is 40. 0 means no limit"""
        if (self.nsxlib and
            not self.nsxlib.feature_supported(
                nsx_constants.FEATURE_RATE_LIMIT)):
            msg = (_("Rate limit is not supported by NSX version %s") %
                   self.nsxlib.get_version())
            raise exceptions.ManagerError(details=msg)

        properties = self.get_properties()
        if 'service_properties' in properties:
            properties['service_properties'][
                'client_api_rate_limit'] = int(value)

        # update the value using a PUT command, which is expected to return 202
        expected_results = [requests.codes.accepted]
        self.client.update(self.uri_segment, properties,
                           expected_results=expected_results)

        # restart the http service using POST, which is expected to return 202
        restart_url = self.uri_segment + '?action=restart'
        self.client.create(restart_url, expected_results=expected_results)

    def delete(self, uuid):
        """Not supported"""
        msg = _("Delete is not supported for %s") % self.uri_segment
        raise exceptions.ManagerError(details=msg)

    def get(self, uuid):
        """Not supported"""
        msg = _("Get is not supported for %s") % self.uri_segment
        raise exceptions.ManagerError(details=msg)

    def list(self):
        """Not supported"""
        msg = _("List is not supported for %s") % self.uri_segment
        raise exceptions.ManagerError(details=msg)

    def find_by_display_name(self, display_name):
        """Not supported"""
        msg = _("Find is not supported for %s") % self.uri_segment
        raise exceptions.ManagerError(details=msg)


class NsxlibClusterNodesConfig(utils.NsxLibApiBase):
    @property
    def uri_segment(self):
        return 'cluster/nodes'

    @property
    def resource_type(self):
        return 'ClusterNodeConfig'

    def delete(self, uuid):
        """Not supported"""
        msg = _("Delete is not supported for %s") % self.uri_segment
        raise exceptions.ManagerError(details=msg)

    def get_managers_ips(self):
        manager_ips = []
        nodes_config = self.client.get(self.get_path())['results']
        for node in nodes_config:
            if 'manager_role' in node:
                manager_conf = node['manager_role']
                if 'api_listen_addr' in manager_conf:
                    list_addr = manager_conf['mgmt_cluster_listen_addr']
                    ip = list_addr['ip_address']
                    if ip != '127.0.0.1':
                        manager_ips.append(list_addr['ip_address'])

        return manager_ips


class NsxlibHostSwitchProfiles(utils.NsxLibApiBase):
    @property
    def uri_segment(self):
        return 'host-switch-profiles'

    @property
    def resource_type(self):
        return 'UplinkHostSwitchProfile'


class Inventory(utils.NsxLibApiBase):
    """REST APIs to support inventory service."""

    RESOURCES_PATH = {"ContainerCluster": "container-clusters",
                      "ContainerProject": "container-projects",
                      "ContainerApplication": "container-applications",
                      "ContainerApplicationInstance":
                      "container-application-instances",
                      "ContainerClusterNode": "container-cluster-nodes",
                      "ContainerNetworkPolicy": "container-network-policies",
                      "ContainerIngressPolicy": "container-ingress-policies"}
    SEGMENT_URI = 'fabric'
    SEGMENT_URI_FOR_UPDATE = 'inventory/container'

    @property
    def uri_segment(self):
        # only serves for get, list and delete. For batch update,
        # another base url is used. The reason is update API is internal.
        return self.SEGMENT_URI

    @property
    def resource_type(self):
        return 'Inventory'

    def update(self, cluster_id, updates):
        """This method supports multiple updates in a batching way."""
        items = []
        for update_type, update_object in updates:
            item = {}
            item["object_update_type"] = update_type
            item["container_object"] = update_object
            items.append(item)
        body = {"container_inventory_objects": items}
        request_url = "%s/%s?action=updates" % (
            self.SEGMENT_URI_FOR_UPDATE, cluster_id)
        return self.client.url_post(request_url, body)

    def get(self, resource_type, resource_id):
        if not resource_type:
            msg = "null resource type is not supported"
            raise exceptions.ResourceNotFound(details=msg)
        request_url = "%s/%s" % (
            self.get_path(self._get_path_for_resource(resource_type)),
            resource_id)
        return self.client.url_get(request_url)

    def list(self, cluster_id, resource_type):
        if not resource_type:
            msg = "null resource type is not supported"
            raise exceptions.ResourceNotFound(details=msg)
        request_url = "%s?container_cluster_id=%s" % (
            self.get_path(self._get_path_for_resource(resource_type)),
            cluster_id)
        return self.client.url_list(request_url)

    def delete(self, resource_type, resource_id):
        if not resource_type:
            msg = "null resource type is not supported"
            raise exceptions.ResourceNotFound(details=msg)
        request_url = "%s/%s" % (
            self.get_path(self._get_path_for_resource(resource_type)),
            resource_id)
        return self.client.url_delete(request_url)

    def create(self, resource_type, resource):
        if not resource_type:
            msg = "null resource type is not supported"
            raise exceptions.ResourceNotFound(details=msg)
        request_url = self.get_path(
            self._get_path_for_resource(resource_type))
        return self.client.url_post(request_url, resource)

    def _get_path_for_resource(self, resource_type):
        path = self.RESOURCES_PATH.get(resource_type)
        if not path:
            msg = "backend resource %s is not supported" % resource_type
            raise exceptions.ResourceNotFound(details=msg)
        return path


class SystemHealth(utils.NsxLibApiBase):
    @property
    def uri_segment(self):
        return 'systemhealth'

    @property
    def resource_type(self):
        return 'SystemHealth'

    def create_ncp_status(self, cluster_id, status):
        url = '/container-cluster/ncp/status'
        body = {'cluster_id': cluster_id, 'status': status}
        return self.client.create(self.get_path(url), body=body)



# Copyright 2016  VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log

from requests import exceptions as requests_exceptions

from vmware_nsxlib.v3 import exceptions as v3_exceptions

LOG = log.getLogger(__name__)


class ExceptionConfig(object):

    def __init__(self):
        # When hit during API call, these exceptions will mark
        # endpoint as DOWN immediately
        # This setting has no effect on keepalive validation
        self.ground_triggers = [requests_exceptions.ConnectionError,
                                requests_exceptions.Timeout]

        # When hit during API call, these exceptions will be
        # retried with next available endpoint
        # When hit during validation, these exception will not
        # mark endpoint as DOWN
        self.retriables = [v3_exceptions.APITransactionAborted,
                           v3_exceptions.CannotConnectToServer,
                           v3_exceptions.ServerBusy]

        # When hit during API call, these exceptions will be retried
        # after the endpoints are regenerated with up-to-date auth
        # credentials / tokens
        self.regenerate_triggers = [v3_exceptions.InvalidCredentials,
                                    v3_exceptions.ClientCertificateNotTrusted,
                                    v3_exceptions.BadXSRFToken]

    def should_ground_endpoint(self, ex):
        for exception in self.ground_triggers:
            if isinstance(ex, exception):
                return True

        return False

    def should_retry(self, ex):
        for exception in self.retriables:
            if isinstance(ex, exception):
                return True

        return False

    def should_regenerate(self, ex):
        for exception in self.regenerate_triggers:
            if isinstance(ex, exception):
                return True

        return False


class NsxLibConfig(object):
    """Class holding all the configuration parameters used by the nsxlib code.

    :param nsx_api_managers: List of IP addresses of the NSX managers.
                             Each IP address should be of the form:
                             [<scheme>://]<ip_address>[:<port>]
                             If scheme is not provided https is used.
                             If port is not provided port 80 is used for http
                             and port 443 for https.
    :param username: User name for the NSX manager
    :param password: Password for the NSX manager
    :param client_cert_provider: None, or ClientCertProvider object.
                             If specified, nsxlib will use client cert auth
                             instead of basic authentication.
    :param insecure: If true, the NSX Manager server certificate is not
                     verified. If false the CA bundle specified via "ca_file"
                     will be used or if unset the "thumbprint" will be used.
                     If "thumbprint" is unset, the default system root CAs
                     will be used.
    :param ca_file: Specify a CA bundle file to use in verifying the NSX
                    Manager server certificate. This option is ignored if
                    "insecure" is set to True. If "insecure" is set to False
                    and "ca_file" is unset, the "thumbprint" will be used.
                    If "thumbprint" is unset, the system root CAs will be
                    used to verify the server certificate.
    :param thumbprint: Specify a thumbprint string to use in verifying the
                       NSX Manager server certificate. This option is ignored
                       if "insecure" is set to True or "ca_file" is defined.
    :param token_provider: None, or instance of implemented AbstractJWTProvider
                           which will return the JSON Web Token used in the
                           requests in NSX for authorization.

    :param concurrent_connections: Maximum concurrent connections to each NSX
                                   manager.
    :param retries: Maximum number of times to retry a HTTP connection.
    :param http_timeout: The time in seconds before aborting a HTTP connection
                         to a NSX manager.
    :param http_read_timeout: The time in seconds before aborting a HTTP read
                              response from a NSX manager.
    :param conn_idle_timeout: The amount of time in seconds to wait before
                              ensuring connectivity to the NSX manager if no
                              manager connection has been used.
    :param http_provider: HTTPProvider object, or None.

    :param max_attempts: Maximum number of times to retry API requests upon
                         stale revision errors.

    :param plugin_scope: The default scope for the v3 api-version tag
    :param plugin_tag: The value for the v3 api-version tag
    :param plugin_ver: The version of the plugin used as the 'os-api-version'
                       tag value in the v3 api-version tag
    :param dns_nameservers: List of nameservers to configure for the DHCP
                            binding entries. These will be used if there are
                            no nameservers defined on the subnet.
    :param dns_domain: Domain to use for building the hostnames.
    :param allow_overwrite_header: If True, a default header of
                                   X-Allow-Overwrite:true will be added to all
                                   the requests, to allow admin user to update/
                                   delete all entries.
    :param rate_limit_retry: If True, the client will retry requests failed on
           "Too many requests" error.
    :param cluster_unavailable_retry: If True, skip fatal errors when no
                                      endpoint in the NSX management cluster is
                                      available to serve a request, and retry
                                      the request instead. This setting can
                                      not be False if single endpoint is
                                      configured in the cluster, since there
                                      will be no keepalive probes in this
                                      case.
    :param api_rate_limit_per_endpoint: If set to positive integer, API calls
                                        sent to each endpoint will be limited
                                        to a max rate of this value per second.
                                        The rate limit is not enforced on
                                        connection validations. This option
                                        defaults to None, which disables rate
                                        limit.
    :param api_rate_mode: Algorithm used to adaptively adjust max API rate
                          limit. If not set, the max rate will not be
                          automatically changed. If set to 'AIMD', max API
                          rate will be increase by 1 after successful calls
                          that was blocked before sent, and will be decreased
                          by half after 429/503 error for each period.
                          The rate has hard max limit of min(100/s, param
                          api_rate_limit_per_endpoint).
    :param api_log_mode: Option to collect API call logs within nsxlib.
                         When set to API_LOG_PER_CLUSTER, API calls sent to all
                         endpoints will be collected at one place.
                         When set to API_LOG_PER_ENDPOINT, API calls sent to
                         each endpoint will be collected individually.
                         By default, this option is disabled as set to None.
    :param enable_health_check: Options to enable or disable health check for
                                all endpoints when initializing cluster API.
                                The checking including endpoint connection
                                validation and health check loop.
                                For some condition, eg election process. It
                                does not need to check the endpoint's
                                accessibility.
                                By default, this option is set to True.

    -- Additional parameters which are relevant only for the Policy manager:
    :param allow_passthrough: If True, use nsx manager api for cases which are
                              not supported by the policy manager api.
    :param realization_max_attempts: Maximum number of times to retry while
                                     waiting for a resource to be realized.
    :param realization_wait_sec: Number of seconds to wait between attempts
                                 for a resource to be realized.
    """

    def __init__(self,
                 nsx_api_managers=None,
                 username=None,
                 password=None,
                 client_cert_provider=None,
                 insecure=True,
                 ca_file=None,
                 thumbprint=None,
                 token_provider=None,
                 concurrent_connections=10,
                 retries=3,
                 http_timeout=10,
                 http_read_timeout=180,
                 conn_idle_timeout=10,
                 http_provider=None,
                 max_attempts=10,
                 plugin_scope=None,
                 plugin_tag=None,
                 plugin_ver=None,
                 dns_nameservers=None,
                 dns_domain='openstacklocal',
                 allow_overwrite_header=False,
                 rate_limit_retry=True,
                 cluster_unavailable_retry=False,
                 allow_passthrough=False,
                 realization_max_attempts=50,
                 realization_wait_sec=1.0,
                 api_rate_limit_per_endpoint=None,
                 api_rate_mode=None,
                 exception_config=None,
                 api_log_mode=None,
                 enable_health_check=True):

        self.nsx_api_managers = nsx_api_managers
        self._username = username
        self._password = password
        self._ca_file = ca_file
        self._thumbprint = thumbprint
        self.insecure = insecure
        self.concurrent_connections = concurrent_connections
        self.retries = retries
        self.http_timeout = http_timeout
        self.http_read_timeout = http_read_timeout
        self.conn_idle_timeout = conn_idle_timeout
        self.http_provider = http_provider
        self.client_cert_provider = client_cert_provider
        self.token_provider = token_provider
        self.max_attempts = max_attempts
        self.plugin_scope = plugin_scope
        self.plugin_tag = plugin_tag
        self.plugin_ver = plugin_ver
        self.dns_nameservers = dns_nameservers or []
        self.dns_domain = dns_domain
        self.allow_overwrite_header = allow_overwrite_header
        self.rate_limit_retry = rate_limit_retry
        self.cluster_unavailable_retry = cluster_unavailable_retry
        self.allow_passthrough = allow_passthrough
        self.realization_max_attempts = realization_max_attempts
        self.realization_wait_sec = realization_wait_sec
        self.api_rate_limit_per_endpoint = api_rate_limit_per_endpoint
        self.api_rate_mode = api_rate_mode
        self.exception_config = exception_config or ExceptionConfig()
        self.api_log_mode = api_log_mode
        self.enable_health_check = enable_health_check

        if len(nsx_api_managers) == 1 and not self.cluster_unavailable_retry:
            LOG.warning("When only one endpoint is provided, keepalive probes"
                        " are disabled. For the system to be able to recover"
                        " from DOWN state, cluster_unavailable_retry is set"
                        " to True, overriding provided configuration")
            self.cluster_unavailable_retry = True

        if len(nsx_api_managers) > self.max_attempts:
            LOG.warning("max_attempts setting (%d) is lower than amount of"
                        " endpoints (%d), which means that not all endpoints"
                        " will be probed in case of retriable error",
                        self.max_attempts, len(nsx_api_managers))

    def extend(self, keepalive_section, validate_connection_method=None,
               url_base=None):
        if keepalive_section or validate_connection_method:
            LOG.warning("keepalive_section and validate_connection_method are"
                        " no longer used to conduct keepalive probes. For"
                        " most efficient keepalive roundtrip, proxy health"
                        " API is always used.")
        self.url_base = url_base

    def _attribute_by_index(self, scalar_or_list, index):
        if isinstance(scalar_or_list, list):
            if not len(scalar_or_list):
                return None
            if len(scalar_or_list) > index:
                return scalar_or_list[index]
            # if not long enough - use the first one as default
            return scalar_or_list[0]
        # this is a scalar
        return scalar_or_list

    def username(self, index):
        return self._attribute_by_index(self._username, index)

    def password(self, index):
        return self._attribute_by_index(self._password, index)

    def ca_file(self, index):
        return self._attribute_by_index(self._ca_file, index)

    def thumbprint(self, index):
        return self._attribute_by_index(self._thumbprint, index)



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from oslo_log import log
from oslo_log import versionutils

from vmware_nsxlib._i18n import _
from vmware_nsxlib.v3 import exceptions
from vmware_nsxlib.v3 import nsx_constants
from vmware_nsxlib.v3 import utils

LOG = log.getLogger(__name__)

SwitchingProfileTypeId = collections.namedtuple(
    'SwitchingProfileTypeId', 'profile_type, profile_id')

PacketAddressClassifier = collections.namedtuple(
    'PacketAddressClassifier', 'ip_address, mac_address, vlan')


class NsxLibPortMirror(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'mirror-sessions'

    @property
    def resource_type(self):
        return 'PortMirroringSession'

    def create_session(self, source_ports, dest_ports, direction,
                       description, name, tags):
        """Create a PortMirror Session on the backend.

        :param source_ports: List of UUIDs of the ports whose traffic is to be
                            mirrored.
        :param dest_ports: List of UUIDs of the ports where the mirrored
                          traffic is to be sent.
        :param direction: String representing the direction of traffic to be
                          mirrored. [INGRESS, EGRESS, BIDIRECTIONAL]
        :param description: String representing the description of the session.
        :param name: String representing the name of the session.
        :param tags: nsx backend specific tags.
        """

        body = {'direction': direction,
                'tags': tags,
                'display_name': name,
                'description': description,
                'mirror_sources': source_ports,
                'mirror_destination': dest_ports}
        return self.client.create(self.get_path(), body)

    def delete_session(self, mirror_session_id):
        """Delete a PortMirror session on the backend.

        :param mirror_session_id: string representing the UUID of the port
                                  mirror session to be deleted.
        """
        self._delete_with_retry(mirror_session_id)


class NsxLibBridgeEndpointProfile(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'bridge-endpoint-profiles'

    @property
    def resource_type(self):
        return 'BridgeEndpointProfile'

    def create(self, display_name, edge_cluster_id, tags,
               edge_cluster_member_indexes=None, failover_mode=None):
        """Create a bridge endpoint profile on the backend.

        Create a bridge endpoint profile for a given edge cluster.
        :param display_name: name of the bridge endpoint profile
        :param edge_cluster_id: identifier of the edge cluster this profile
                                should be associated with.
        :param tags: tags for the newly created resource.
        :param edge_cluster_member_indexes: iterable of integers specifying
                                            edge cluster members where the
                                            bridge endpoints will be created
        :param failover_mode: failover mode for the profile. Could be either
                              PREEMPTIVE or NON_PREEMPTIVE.
        """
        tags = tags or []
        body = {'display_name': display_name,
                'tags': tags}
        if failover_mode:
            body['failover_mode'] = failover_mode
        if edge_cluster_member_indexes:
            # Test for a list of integers
            try:
                member_indexes = [int(member_idx) for member_idx in
                                  edge_cluster_member_indexes]
                body['edge_cluster_member_indexes'] = member_indexes
            except (TypeError, ValueError) as e:
                LOG.Error("Invalid values for member indexes: %s", e)
                raise exceptions.InvalidInput(
                    operation='Create BridgeEndpointProfile',
                    arg_val=edge_cluster_member_indexes,
                    arg_name='edge_cluster_member_indexes')

        return self.client.create(self.get_path(), body)

    def delete(self, bridge_endpoint_profile_id):
        """Delete a bridge endpoint profile on the backend.

        :param bridge_endpoint_profile_id: string representing the UUID of
                                           the bridge endpoint profile to be
                                           deleted.
        """
        self._delete_with_retry(bridge_endpoint_profile_id)


class NsxLibBridgeEndpoint(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'bridge-endpoints'

    @property
    def resource_type(self):
        return 'BridgeEndpoint'

    def create(self, device_name, vlan_transport_zone_id, vlan_id, tags):
        """Create a bridge endpoint on the backend.

        Create a bridge endpoint resource on a bridge cluster for the L2
        gateway network connection.
        :param device_name: device_name actually refers to the bridge cluster's
                            UUID.
        :param vlan_transport_zone_id: identifier of the transport zone id
                                       where the endpoint will be created.
                                       Mandatory for endpoints on edge
                                       clusters.
        :param vlan_id: integer representing the VLAN segmentation ID.
        :param tags: nsx backend specific tags.
        """
        body = {'bridge_endpoint_profile_id': device_name,
                'vlan_transport_zone_id': vlan_transport_zone_id,
                'tags': tags,
                'vlan': vlan_id}
        return self.client.create(self.get_path(), body)

    def delete(self, bridge_endpoint_id):
        """Delete a bridge endpoint on the backend.

        :param bridge_endpoint_id: string representing the UUID of the bridge
                                   endpoint to be deleted.
        """
        self._delete_with_retry(bridge_endpoint_id)


class NsxLibLogicalSwitch(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'logical-switches'

    @property
    def resource_type(self):
        return 'LogicalSwitch'

    def create(self, display_name, transport_zone_id, tags,
               replication_mode=nsx_constants.MTEP,
               admin_state=True, vlan_id=None, ip_pool_id=None,
               mac_pool_id=None, description=None,
               trunk_vlan_range=None):
        operation = "Create logical switch"
        if display_name:
            display_name = utils.escape_display_name(display_name)
        # TODO(salv-orlando): Validate Replication mode and admin_state
        # NOTE: These checks might be moved to the API client library if one
        # that performs such checks in the client is available
        body = {'transport_zone_id': transport_zone_id,
                'replication_mode': replication_mode,
                'display_name': display_name,
                'tags': tags}

        if admin_state:
            body['admin_state'] = nsx_constants.ADMIN_STATE_UP
        else:
            body['admin_state'] = nsx_constants.ADMIN_STATE_DOWN

        if trunk_vlan_range:
            failed = False
            if (self.nsxlib and
                self.nsxlib.feature_supported(
                    nsx_constants.FEATURE_TRUNK_VLAN)):
                if vlan_id is not None:
                    failed = True
                    LOG.error("Failed to create logical switch %(name)s with "
                              "trunk vlan: vlan id %(vlan)s is used.",
                              {'name': display_name, 'vlan': vlan_id})
                elif (len(trunk_vlan_range) != 2 or
                      trunk_vlan_range[0] > trunk_vlan_range[1]):
                    failed = True
                    LOG.error("Failed to create logical switch %(name)s with "
                              "trunk vlan: illegal range (%(trunk)s) is used.",
                              {'name': display_name,
                               'trunk': trunk_vlan_range})
                else:
                    body['vlan_trunk_spec'] = {'vlan_ranges': [
                        {'start': trunk_vlan_range[0],
                         'end': trunk_vlan_range[1]}]}
            else:
                LOG.error("Failed to create logical switch %s with trunk "
                          "vlan: this feature is not supported.", display_name)
                failed = True
            if failed:
                raise exceptions.InvalidInput(
                    operation=operation,
                    arg_val=trunk_vlan_range,
                    arg_name='trunk_vlan_range')
        elif vlan_id:
            body['vlan'] = vlan_id

        if ip_pool_id:
            body['ip_pool_id'] = ip_pool_id

        if mac_pool_id:
            body['mac_pool_id'] = mac_pool_id

        if description is not None:
            body['description'] = description

        return self.client.create(self.get_path(), body)

    def delete(self, lswitch_id):
        resource = '%s?detach=true&cascade=true' % lswitch_id
        self._delete_with_retry(resource)

    def update(self, lswitch_id, name=None, admin_state=None, tags=None,
               description=None):
        body = {}
        if name:
            name = utils.escape_display_name(name)
            body['display_name'] = name
        if admin_state is not None:
            if admin_state:
                body['admin_state'] = nsx_constants.ADMIN_STATE_UP
            else:
                body['admin_state'] = nsx_constants.ADMIN_STATE_DOWN
        if tags is not None:
            body['tags'] = tags
        if description is not None:
            body['description'] = description
        return self._update_with_retry(lswitch_id, body)


class SwitchingProfileTypes(object):
    IP_DISCOVERY = 'IpDiscoverySwitchingProfile'
    MAC_LEARNING = 'MacManagementSwitchingProfile'
    PORT_MIRRORING = 'PortMirroringSwitchingProfile'
    QOS = 'QosSwitchingProfile'
    SPOOF_GUARD = 'SpoofGuardSwitchingProfile'
    SWITCH_SECURITY = 'SwitchSecuritySwitchingProfile'


class WhiteListAddressTypes(object):
    PORT = 'LPORT_BINDINGS'
    SWITCH = 'LSWITCH_BINDINGS'


class NsxLibSwitchingProfile(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'switching-profiles'

    def get_path(self, resource=None, query_params=None):
        path = super().get_path(resource)
        if query_params:
            param_str = "&".join("%s=%s" % (k, v) for (k, v) in
                                 query_params.items())
            path = '%s?%s' % (path, param_str)
        return path

    def list(self):
        return self.client.list(
            self.get_path(query_params={'include_system_owned': True}))

    def create(self, profile_type, display_name=None,
               description=None, **api_args):
        body = {
            'resource_type': profile_type,
            'display_name': display_name or '',
            'description': description or ''
        }
        body.update(api_args)

        return self.client.create(self.get_path(), body=body)

    def update(self, uuid, profile_type, **api_args):
        body = {
            'resource_type': profile_type
        }
        body.update(api_args)

        return self.client.update(self.get_path(uuid), body=body)

    def create_spoofguard_profile(self, display_name,
                                  description,
                                  whitelist_ports=False,
                                  whitelist_switches=False,
                                  tags=None):
        whitelist_providers = []
        if whitelist_ports:
            whitelist_providers.append(WhiteListAddressTypes.PORT)
        if whitelist_switches:
            whitelist_providers.append(WhiteListAddressTypes.SWITCH)

        return self.create(SwitchingProfileTypes.SPOOF_GUARD,
                           display_name=display_name,
                           description=description,
                           white_list_providers=whitelist_providers,
                           tags=tags or [])

    def create_dhcp_profile(self, display_name,
                            description, tags=None):
        dhcp_filter = {
            'client_block_enabled': True,
            'server_block_enabled': False
        }
        rate_limits = {
            'enabled': False,
            'rx_broadcast': 0,
            'tx_broadcast': 0,
            'rx_multicast': 0,
            'tx_multicast': 0
        }
        bpdu_filter = {
            'enabled': True,
            'white_list': []
        }
        return self.create(SwitchingProfileTypes.SWITCH_SECURITY,
                           display_name=display_name,
                           description=description,
                           tags=tags or [],
                           dhcp_filter=dhcp_filter,
                           rate_limits=rate_limits,
                           bpdu_filter=bpdu_filter,
                           block_non_ip_traffic=True)

    def create_mac_learning_profile(self, display_name, description,
                                    mac_learning_enabled=True,
                                    tags=None):
        mac_learning = {
            'enabled': mac_learning_enabled,
            'unicast_flooding_allowed': mac_learning_enabled
        }
        return self.create(SwitchingProfileTypes.MAC_LEARNING,
                           display_name=display_name,
                           description=description,
                           tags=tags or [],
                           mac_learning=mac_learning,
                           mac_change_allowed=True)

    def create_port_mirror_profile(self, display_name, description,
                                   direction, destinations, tags=None):
        return self.create(SwitchingProfileTypes.PORT_MIRRORING,
                           display_name=display_name,
                           description=description,
                           tags=tags or [],
                           direction=direction,
                           destinations=destinations)

    @classmethod
    def build_switch_profile_ids(cls, client, *profiles):
        ids = []
        for profile in profiles:
            if isinstance(profile, str):
                profile = client.get(profile)
            if not isinstance(profile, SwitchingProfileTypeId):
                profile = SwitchingProfileTypeId(
                    profile.get('key', profile.get('resource_type')),
                    profile.get('value', profile.get('id')))
            ids.append(profile)
        return ids


class NsxLibQosSwitchingProfile(NsxLibSwitchingProfile):

    @property
    def resource_type(self):
        return 'QosSwitchingProfile'

    def _build_args(self, tags, name=None, description=None):
        body = {"resource_type": "QosSwitchingProfile",
                "tags": tags}
        return self._update_args(
            body, name=name, description=description)

    def _update_args(self, body, name=None, description=None):
        if name:
            body["display_name"] = name
        if description:
            body["description"] = description
        return body

    def _get_resource_type(self, direction):
        if direction == nsx_constants.EGRESS:
            return nsx_constants.EGRESS_SHAPING
        return nsx_constants.INGRESS_SHAPING

    def _enable_shaping_in_args(self, body, burst_size=None,
                                peak_bandwidth=None, average_bandwidth=None,
                                direction=None):
        resource_type = self._get_resource_type(direction)
        for shaper in body["shaper_configuration"]:
            if shaper["resource_type"] == resource_type:
                shaper["enabled"] = True
                if burst_size is not None:
                    shaper["burst_size_bytes"] = burst_size
                if peak_bandwidth is not None:
                    shaper["peak_bandwidth_mbps"] = peak_bandwidth
                if average_bandwidth is not None:
                    shaper["average_bandwidth_mbps"] = average_bandwidth
                break

        return body

    def _disable_shaping_in_args(self, body, direction=None):
        resource_type = self._get_resource_type(direction)
        for shaper in body["shaper_configuration"]:
            if shaper["resource_type"] == resource_type:
                shaper["enabled"] = False
                shaper["burst_size_bytes"] = 0
                shaper["peak_bandwidth_mbps"] = 0
                shaper["average_bandwidth_mbps"] = 0
                break

        return body

    def _update_dscp_in_args(self, body, qos_marking, dscp):
        body["dscp"] = {}
        body["dscp"]["mode"] = qos_marking.upper()
        if dscp:
            body["dscp"]["priority"] = dscp

        return body

    def create(self, tags, name=None, description=None):
        body = self._build_args(tags, name, description)
        return self.client.create(self.get_path(), body)

    def update(self, profile_id, tags, name=None, description=None):
        # update the relevant fields
        body = {}
        body = self._update_args(body, name, description)
        if tags is not None:
            body['tags'] = tags
        return self._update_with_retry(profile_id, body)

    def update_shaping(self, profile_id,
                       shaping_enabled=False,
                       burst_size=None,
                       peak_bandwidth=None,
                       average_bandwidth=None,
                       qos_marking=None, dscp=None,
                       direction=nsx_constants.INGRESS):
        versionutils.report_deprecated_feature(
            LOG,
            'NsxLibQosSwitchingProfile.update_shaping is deprecated. '
            'Please use set_profile_shaping instead.')
        # get the current configuration
        body = self.get(profile_id)
        # update the relevant fields
        if shaping_enabled:
            body = self._enable_shaping_in_args(
                body, burst_size=burst_size,
                peak_bandwidth=peak_bandwidth,
                average_bandwidth=average_bandwidth,
                direction=direction)
        else:
            body = self._disable_shaping_in_args(body, direction=direction)
        body = self._update_dscp_in_args(body, qos_marking, dscp)
        return self._update_with_retry(profile_id, body)

    def set_profile_shaping(self, profile_id,
                            ingress_bw_enabled=False,
                            ingress_burst_size=None,
                            ingress_peak_bandwidth=None,
                            ingress_average_bandwidth=None,
                            egress_bw_enabled=False,
                            egress_burst_size=None,
                            egress_peak_bandwidth=None,
                            egress_average_bandwidth=None,
                            qos_marking='trusted', dscp=None):
        """Set all shaping parameters in the QoS switch profile"""
        # get the current configuration
        body = self.get(profile_id)

        # update the ingress shaping
        if ingress_bw_enabled:
            body = self._enable_shaping_in_args(
                body, burst_size=ingress_burst_size,
                peak_bandwidth=ingress_peak_bandwidth,
                average_bandwidth=ingress_average_bandwidth,
                direction=nsx_constants.INGRESS)
        else:
            body = self._disable_shaping_in_args(
                body, direction=nsx_constants.INGRESS)

        # update the egress shaping
        if egress_bw_enabled:
            body = self._enable_shaping_in_args(
                body, burst_size=egress_burst_size,
                peak_bandwidth=egress_peak_bandwidth,
                average_bandwidth=egress_average_bandwidth,
                direction=nsx_constants.EGRESS)
        else:
            body = self._disable_shaping_in_args(
                body, direction=nsx_constants.EGRESS)

        # update dscp marking
        body = self._update_dscp_in_args(body, qos_marking, dscp)

        # update the profile in the backend
        return self._update_with_retry(profile_id, body)


class NsxLibLogicalRouter(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'logical-routers'

    @property
    def resource_type(self):
        return 'LogicalRouter'

    def _delete_resource_by_values(self, resource,
                                   skip_not_found=True,
                                   strict_mode=True,
                                   **kwargs):
        """Delete resource objects matching the values in kwargs

        If skip_not_found is True - do not raise an exception if no object was
        found.
        If strict_mode is True - warnings will be issued if 0 or >1 objects
        where deleted.
        """
        resources_list = self.client.list(resource)
        matched_num = 0
        for res in resources_list['results']:
            if utils.dict_match(kwargs, res):
                LOG.debug("Deleting %s from resource %s", res, resource)
                delete_resource = resource + "/" + str(res['id'])
                self._delete_by_path_with_retry(delete_resource)
                matched_num = matched_num + 1
        if matched_num == 0:
            if skip_not_found:
                if strict_mode:
                    LOG.warning("No resource in %(res)s matched for values: "
                                "%(values)s", {'res': resource,
                                               'values': kwargs})
            else:
                err_msg = (_("No resource in %(res)s matched for values: "
                             "%(values)s") % {'res': resource,
                                              'values': kwargs})
                raise exceptions.ResourceNotFound(
                    manager=self.client.nsx_api_managers,
                    operation=err_msg)
        elif matched_num > 1 and strict_mode:
            LOG.warning("%(num)s resources in %(res)s matched for values: "
                        "%(values)s", {'num': matched_num,
                                       'res': resource,
                                       'values': kwargs})

    def _validate_nat_rule_action(self, action):
        if not action:
            return
        if action in ['SNAT', 'DNAT', 'NO_NAT', 'REFLEXIVE']:
            # legal values for all NSX versions
            return
        if (action not in ['NO_SNAT', 'NO_DNAT'] or (
            self.nsxlib and not self.nsxlib.feature_supported(
                nsx_constants.FEATURE_NO_DNAT_NO_SNAT))):
            raise exceptions.InvalidInput(
                operation="Create/Update NAT rule",
                arg_val=action,
                arg_name='action')

    def add_nat_rule(self, logical_router_id, action, translated_network,
                     source_net=None, dest_net=None,
                     enabled=True, rule_priority=None,
                     match_ports=None, match_protocol=None,
                     match_resource_type=None,
                     bypass_firewall=True, logging=None,
                     tags=None,
                     display_name=None):
        self._validate_nat_rule_action(action)
        resource = 'logical-routers/%s/nat/rules' % logical_router_id
        body = {'action': action,
                'enabled': enabled,
                'translated_network': translated_network}
        if source_net:
            body['match_source_network'] = source_net
        if dest_net:
            body['match_destination_network'] = dest_net
        if rule_priority:
            body['rule_priority'] = rule_priority
        if match_ports:
            body['match_service'] = {
                'resource_type': (match_resource_type or
                                  nsx_constants.L4_PORT_SET_NSSERVICE),
                'destination_ports': match_ports,
                'l4_protocol': match_protocol or nsx_constants.TCP}

        # nat_pass parameter is supported with the router firewall feature
        if (self.nsxlib and
            self.nsxlib.feature_supported(
                nsx_constants.FEATURE_ROUTER_FIREWALL)):
            body['nat_pass'] = bypass_firewall
        elif not bypass_firewall:
            LOG.error("Ignoring bypass_firewall for router %s nat rule: "
                      "this feature is not supported.", logical_router_id)
        if tags is not None:
            body['tags'] = tags
        if display_name:
            body['display_name'] = display_name
        if logging is not None:
            body['logging'] = logging
        return self.client.create(resource, body)

    def change_edge_firewall_status(self, logical_router_id, action):
        resource = 'firewall/status/logical_routers/%s?action=%s' % (
            logical_router_id, action)
        return self.client.create(resource)

    def add_static_route(self, logical_router_id, dest_cidr, nexthop,
                         tags=None):
        resource = ('logical-routers/%s/routing/static-routes' %
                    logical_router_id)
        body = {}
        if dest_cidr:
            body['network'] = dest_cidr
        if nexthop:
            body['next_hops'] = [{"ip_address": nexthop}]
        if tags is not None:
            body['tags'] = tags
        return self.client.create(resource, body)

    def delete_static_route(self, logical_router_id, static_route_id):
        path = 'logical-routers/%s/routing/static-routes/%s' % (
            logical_router_id, static_route_id)
        self._delete_by_path_with_retry(path)

    def delete_static_route_by_values(self, logical_router_id,
                                      dest_cidr=None, nexthop=None):
        resource = ('logical-routers/%s/routing/static-routes' %
                    logical_router_id)
        kwargs = {}
        if dest_cidr:
            kwargs['network'] = dest_cidr
        if nexthop:
            kwargs['next_hops'] = [{"ip_address": nexthop}]
        return self._delete_resource_by_values(resource, **kwargs)

    def list_static_routes(self, logical_router_id):
        resource = ('logical-routers/%s/routing/static-routes' %
                    logical_router_id)
        return self.client.list(resource)

    def delete_nat_rule(self, logical_router_id, nat_rule_id):
        path = 'logical-routers/%s/nat/rules/%s' % (logical_router_id,
                                                    nat_rule_id)
        self._delete_by_path_with_retry(path)

    def delete_nat_rule_by_values(self, logical_router_id,
                                  strict_mode=True,
                                  skip_not_found=True,
                                  **kwargs):
        resource = 'logical-routers/%s/nat/rules' % logical_router_id
        return self._delete_resource_by_values(
            resource,
            skip_not_found=skip_not_found,
            strict_mode=strict_mode,
            **kwargs)

    def list_nat_rules(self, logical_router_id):
        resource = 'logical-routers/%s/nat/rules' % logical_router_id
        return self.client.list(resource)

    def update_nat_rule(self, logical_router_id, nat_rule_id, **kwargs):
        if 'action' in kwargs:
            self._validate_nat_rule_action(kwargs['action'])
        resource = 'logical-routers/%s/nat/rules/%s' % (
            logical_router_id, nat_rule_id)
        return self._update_resource(resource, kwargs, retry=True)

    def update_advertisement(self, logical_router_id, **kwargs):
        resource = ('logical-routers/%s/routing/advertisement' %
                    logical_router_id)
        # ignore load balancing flags if lb is the not supported
        if (self.nsxlib and
            not self.nsxlib.feature_supported(
                nsx_constants.FEATURE_LOAD_BALANCER)):
            for arg in ('advertise_lb_vip', 'advertise_lb_snat_ip'):
                if kwargs[arg]:
                    LOG.error("Ignoring %(arg)s for router %(rtr)s "
                              "update_advertisement: This feature is not "
                              "supported.",
                              {'arg': arg, 'rtr': logical_router_id})
                del kwargs[arg]

        return self._update_resource(resource, kwargs, retry=True)

    def update_advertisement_rules(self, logical_router_id, rules,
                                   name_prefix=None, force=False):
        """Update the router advertisement rules

        If name_prefix is None, replace the entire list of NSX rules with the
        new given 'rules'.
        Else - delete the NSX rules with this name prefix, and add 'rules' to
        the rest.
        """
        resource = ('logical-routers/%s/routing/advertisement/rules' %
                    logical_router_id)
        callback = None

        def update_payload_cbk(revised_payload, requested_payload):
            # delete rules with this prefix:
            new_rules = []
            for rule in revised_payload['rules']:
                if (not rule.get('display_name') or
                    not rule['display_name'].startswith(name_prefix)):
                    new_rules.append(rule)
            # add new rules
            new_rules.extend(requested_payload['rules'])
            revised_payload['rules'] = new_rules
            del requested_payload['rules']

        if name_prefix:
            callback = update_payload_cbk

        # In case of updating advertisement rule on the logical router
        # owned by other Principal Entities, need to force the overwrite
        headers = {'X-Allow-Overwrite': 'true'} if force else None
        return self._update_resource(
            resource, {'rules': rules}, retry=True,
            update_payload_cbk=callback, headers=headers)

    def get_advertisement_rules(self, logical_router_id):
        resource = ('logical-routers/%s/routing/advertisement/rules' %
                    logical_router_id)
        return self.client.get(resource)

    def get_debug_info(self, logical_router_id):
        resource = ('logical-routers/%s/debug-info?format=text' %
                    logical_router_id)
        return self.client.get(resource)

    def get_transportzone_id(self, logical_router_id):
        res = self.get_debug_info(logical_router_id)
        for item in res['componentInfo']:
            if item['componentType'] in (nsx_constants.ROUTER_TYPE_TIER0_DR,
                                         nsx_constants.ROUTER_TYPE_TIER1_DR):
                if item['transportZoneId']:
                    return item['transportZoneId'][0]
        LOG.warning('OverlayTransportZone is not yet available on'
                    ' %s.' % (logical_router_id))

    def create(self, display_name, tags, edge_cluster_uuid=None, tier_0=False,
               description=None, transport_zone_id=None,
               allocation_pool=None, enable_standby_relocation=False,
               failover_mode=None):
        # TODO(salv-orlando): If possible do not manage edge clusters
        # in the main plugin logic.
        router_type = (nsx_constants.ROUTER_TYPE_TIER0 if tier_0 else
                       nsx_constants.ROUTER_TYPE_TIER1)
        body = {'display_name': display_name,
                'router_type': router_type,
                'tags': tags}
        if edge_cluster_uuid:
            body['edge_cluster_id'] = edge_cluster_uuid
        if description:
            body['description'] = description
        if transport_zone_id:
            body['advanced_config'] = {
                'transport_zone_id': transport_zone_id}
        if failover_mode:
            body['failover_mode'] = failover_mode
        allocation_profile = {}
        if allocation_pool:
            allocation_profile['allocation_pool'] = allocation_pool
        if (enable_standby_relocation and self.nsxlib and
                self.nsxlib.feature_supported(
                nsx_constants.FEATURE_ENABLE_STANDBY_RELOCATION)):
            allocation_profile[
                'enable_standby_relocation'] = enable_standby_relocation
        if allocation_profile:
            body['allocation_profile'] = allocation_profile

        return self.client.create(self.get_path(), body=body)

    def delete(self, lrouter_id, force=False):
        url = lrouter_id
        if force:
            url += '?force=%s' % force
        return self._delete_by_path_with_retry(self.get_path(url))

    def update(self, lrouter_id, *args, **kwargs):
        body = {}
        for arg in kwargs:
            # special care for transport_zone_id
            if arg == 'transport_zone_id':
                body['advanced_config'] = {
                    'transport_zone_id': kwargs['transport_zone_id']}
            elif arg == 'enable_standby_relocation':
                if (self.nsxlib and self.nsxlib.feature_supported(
                        nsx_constants.FEATURE_ENABLE_STANDBY_RELOCATION)):
                    body['allocation_profile'] = {
                        'enable_standby_relocation':
                            kwargs['enable_standby_relocation']}
            else:
                body[arg] = kwargs[arg]

        return self._update_with_retry(lrouter_id, body)

    def get_firewall_section_id(self, lrouter_id, router_body=None):
        """Return the id of the auto created firewall section of the router

        If the router was already retrieved from the backend it is possible
        to give it as an input to avoid another backend call.
        In case of multiple sections, the first valid one will be returned
        """
        if not router_body:
            router_body = self.get(lrouter_id)
        if 'firewall_sections' in router_body:
            firewall_sections = router_body['firewall_sections']
            for sec in firewall_sections:
                if (sec.get('is_valid') and
                    sec.get('target_type') == "FirewallSection"):
                    return sec.get('target_id')

    def list(self, router_type=None):
        """List all/by type logical routers."""
        if router_type:
            resource = '%s?router_type=%s' % (self.get_path(), router_type)
        else:
            resource = self.get_path()
        return self.client.list(resource)

    def get_redistribution(self, logical_router_id):
        resource = ('logical-routers/%s/routing/redistribution' %
                    logical_router_id)
        return self.client.get(resource)

    def get_redistribution_rules(self, logical_router_id):
        resource = ('logical-routers/%s/routing/redistribution/rules' %
                    logical_router_id)
        return self.client.get(resource)

    def update_redistribution_rules(self, logical_router_id, rules):
        resource = ('logical-routers/%s/routing/redistribution/rules' %
                    logical_router_id)
        return self._update_resource(resource, {'rules': rules}, retry=True)

    def get_bgp_config(self, logical_router_id):
        resource = ('logical-routers/%s/routing/bgp' % logical_router_id)
        return self.client.get(resource)

    def get_bgp_neighbors(self, logical_router_id):
        resource = ('logical-routers/%s/routing/bgp/neighbors' %
                    logical_router_id)
        return self.client.get(resource)

    def get_route_map(self, logical_router_id, route_map_id):
        resource = ('logical-routers/%s/routing/route-maps/%s' % (
                    logical_router_id, route_map_id))
        return self.client.get(resource)

    def get_ip_prefix_list(self, logical_router_id, ip_prefix_list_id):
        resource = ('logical-routers/%s/routing/ip-prefix-lists/%s' % (
                    logical_router_id, ip_prefix_list_id))
        return self.client.get(resource)


class NsxLibEdgeCluster(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'edge-clusters'

    @property
    def resource_type(self):
        return 'EdgeCluster'

    @property
    def use_cache_for_get(self):
        return True

    def get_transport_nodes(self, uuid):
        ec = self.get(uuid)
        members = []
        for member in ec.get('members', []):
            members.append(member.get('transport_node_id'))
        return members


class NsxLibTransportZone(utils.NsxLibApiBase):

    TRANSPORT_TYPE_VLAN = nsx_constants.TRANSPORT_TYPE_VLAN
    TRANSPORT_TYPE_OVERLAY = nsx_constants.TRANSPORT_TYPE_OVERLAY
    HOST_SWITCH_MODE_ENS = nsx_constants.HOST_SWITCH_MODE_ENS
    HOST_SWITCH_MODE_STANDARD = nsx_constants.HOST_SWITCH_MODE_STANDARD

    @property
    def uri_segment(self):
        return 'transport-zones'

    @property
    def resource_type(self):
        return 'TransportZone'

    @property
    def use_cache_for_get(self):
        return True

    def get_transport_type(self, uuid):
        tz = self.get(uuid)
        return tz['transport_type']

    def get_host_switch_mode(self, uuid):
        tz = self.get(uuid)
        return tz.get('host_switch_mode', self.HOST_SWITCH_MODE_STANDARD)


class NsxLibTransportNode(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'transport-nodes'

    @property
    def resource_type(self):
        return 'TransportNode'

    @property
    def use_cache_for_get(self):
        return True

    def get_transport_zones(self, uuid):
        tn = self.get(uuid)

        if (self.nsxlib and self.nsxlib.feature_supported(
            nsx_constants.FEATURE_GET_TZ_FROM_SWITCH)):
            if (not tn.get('host_switch_spec') or not
                tn['host_switch_spec'].get('host_switches')):
                return []

            host_switches = tn.get('host_switch_spec').get('host_switches', [])
            tzs = []
            for host_switch in host_switches:
                tzs.extend([ep.get('transport_zone_id') for ep in
                            host_switch.get('transport_zone_endpoints', [])])
            return tzs
        else:
            return [ep.get('transport_zone_id') for ep in
                    tn.get('transport_zone_endpoints', [])]


class NsxLibDhcpProfile(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'dhcp/server-profiles'

    @property
    def resource_type(self):
        return 'DhcpProfile'


class NsxLibDhcpRelayService(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'dhcp/relays'

    @property
    def resource_type(self):
        return 'DhcpRelayService'

    @property
    def use_cache_for_get(self):
        return True

    def get_server_ips(self, uuid):
        # Return the server ips of the relay profile attached to this service
        service = self.get(uuid)
        profile_id = service.get('dhcp_relay_profile_id')
        if profile_id and self.nsxlib:
            return self.nsxlib.relay_profile.get_server_ips(profile_id)


class NsxLibDhcpRelayProfile(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'dhcp/relay-profiles'

    @property
    def resource_type(self):
        return 'DhcpRelayProfile'

    @property
    def use_cache_for_get(self):
        return True

    def get_server_ips(self, uuid):
        profile = self.get(uuid)
        return profile.get('server_addresses')


class NsxLibMetadataProxy(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'md-proxies'

    @property
    def resource_type(self):
        return 'MetadataProxy'

    @property
    def use_cache_for_get(self):
        return True

    def update(self, uuid, server_url=None, secret=None, edge_cluster_id=None):
        body = {}
        # update the relevant fields
        if server_url is not None:
            body['metadata_server_url'] = server_url
        if secret is not None:
            body['secret'] = secret
        if edge_cluster_id is not None:
            body['edge_cluster_id'] = edge_cluster_id
        return self._update_with_retry(uuid, body)

    def get_md_proxy_status(self, attachment_id, logical_switch_id):
        """Return all matching logical port statuses"""
        url_suffix = ('/%s/%s/status' %
                      (attachment_id, logical_switch_id))
        return self.client.get(self.get_path(url_suffix))


class NsxLibBridgeCluster(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'bridge-clusters'

    @property
    def resource_type(self):
        return 'BridgeCluster'


class NsxLibIpBlockSubnet(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'pools/ip-subnets'

    @property
    def resource_type(self):
        return 'IpBlockSubnet'

    def create(self, ip_block_id, subnet_size, allow_overwrite=False):
        """Create a IP block subnet on the backend."""
        body = {'size': subnet_size,
                'block_id': ip_block_id}
        headers = None
        if allow_overwrite:
            # In case of manager to policy API resources imports,
            # a Policy owned Manager IpBlock resource might be needed
            # to allocate subnet using Manager APIs.
            headers = {'X-Allow-Overwrite': 'true'}
        return self.client.create(self.get_path(), body, headers=headers)

    def delete(self, subnet_id, allow_overwrite=False):
        """Delete a IP block subnet on the backend."""
        headers = None
        if allow_overwrite:
            # Need to force delete subnet if IpBlock is created by Policy API
            headers = {'X-Allow-Overwrite': 'true'}
        self._delete_with_retry(subnet_id, headers=headers)

    def list(self, ip_block_id):
        resource = '%s?block_id=%s' % (self.get_path(), ip_block_id)
        return self.client.get(resource)


class NsxLibIpBlock(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'pools/ip-blocks'

    @property
    def resource_type(self):
        return 'IpBlock'


class NsxLibFabricVirtualMachine(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'fabric/virtual-machines'

    @property
    def resource_type(self):
        return 'VirtualMachine'

    def get_by_display_name(self, display_name):
        url = '%s?display_name=%s' % (self.get_path(), display_name)
        return self.client.get(url)


class NsxLibFabricVirtualInterface(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'fabric/vifs'

    @property
    def resource_type(self):
        return 'VirtualNetworkInterface'

    def get_by_owner_vm_id(self, owner_vm_id):
        url = '%s?owner_vm_id=%s' % (self.get_path(), owner_vm_id)
        return self.client.get(url)


class NsxLibGlobalRoutingConfig(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'global-configs/RoutingGlobalConfig'

    @property
    def resource_type(self):
        return 'RoutingGlobalConfig'

    def set_l3_forwarding_mode(self, mode):
        config = self.client.get(self.get_path())
        if config['l3_forwarding_mode'] != mode:
            config['l3_forwarding_mode'] = mode
            self.client.update(self.get_path(), config)

    def enable_ipv6(self):
        return self.set_l3_forwarding_mode('IPV4_AND_IPV6')

    def disable_ipv6(self):
        return self.set_l3_forwarding_mode('IPV4_ONLY')
